from unittest.mock import patch, call
import pytest
from time import sleep
import threading
from contextlib import contextmanager, ExitStack
import random
import re

from easypy.concurrency import MultiObject, MultiException, concurrent
from easypy.timing import repeat, timing
from easypy.bunch import Bunch
from easypy.units import Duration

from easypy.sync import iter_wait, wait, iter_wait_progress, Timer, TimeoutException, PredicateNotSatisfied
from easypy.sync import SynchronizationCoordinator, SYNC
from easypy.sync import shared_contextmanager
from easypy.sync import TagAlongThread
from easypy.sync import LoggedRLock, LockLeaseExpired
from easypy.sync import SynchronizedSingleton
from easypy.sync import LoggedCondition

from .test_logging import get_log  # noqa; pytest fixture


def test_shared_contextmanager():

    data = []

    @shared_contextmanager
    def foo(a):
        data.append(a)
        yield a
        data.append(-a)

    with foo(1):
        assert data == [1]

        with foo(2):
            assert data == [1, 2]

            with foo(2):
                assert data == [1, 2]

                with foo(1):
                    assert data == [1, 2]

            assert data == [1, 2]
        assert data == [1, 2, -2]
    assert data == [1, 2, -2, -1]

    data.clear()

    with foo(1):
        assert data == [1]
        with foo(2):
            assert data == [1, 2]
            with foo(2):
                assert data == [1, 2]
            assert data == [1, 2]
        assert data == [1, 2, -2]
    assert data == [1, 2, -2, -1]


def test_shared_contextmanager_method():

    class Foo(object):

        def __init__(self):
            self.data = []

        @shared_contextmanager
        def foo(self, a):
            self.data.append(a)
            yield a
            self.data.append(-a)

    f = Foo()
    g = Foo()

    with f.foo(1), g.foo(5):
        assert f.data == [1]
        assert g.data == [5]

        with f.foo(2), g.foo(6):
            assert f.data == [1, 2]
            assert g.data == [5, 6]

            with f.foo(2), g.foo(6):
                assert f.data == [1, 2]
                assert g.data == [5, 6]

                with f.foo(1), g.foo(5):
                    assert f.data == [1, 2]
                    assert g.data == [5, 6]

            assert f.data == [1, 2]
            assert g.data == [5, 6]
        assert f.data == [1, 2, -2]
        assert g.data == [5, 6, -6]
    assert f.data == [1, 2, -2, -1]
    assert g.data == [5, 6, -6, -5]

    f.data.clear()
    g.data.clear()

    with f.foo(1), g.foo(5):
        assert f.data == [1]
        assert g.data == [5]

        with f.foo(2), g.foo(6):
            assert f.data == [1, 2]
            assert g.data == [5, 6]

            with f.foo(2), g.foo(6):
                assert f.data == [1, 2]
                assert g.data == [5, 6]

                with f.foo(1), g.foo(5):
                    assert f.data == [1, 2]
                    assert g.data == [5, 6]

            assert f.data == [1, 2]
            assert g.data == [5, 6]
        assert f.data == [1, 2, -2]
        assert g.data == [5, 6, -6]
    assert f.data == [1, 2, -2, -1]
    assert g.data == [5, 6, -6, -5]


def test_shared_contextmanager_method_does_not_keep_object_alive_after_done():
    import weakref
    import gc

    class Foo:
        @shared_contextmanager
        def foo(self):
            yield

    foo = Foo()
    weak_foo = weakref.ref(foo)
    with foo.foo():
        del foo
        gc.collect()
        assert weak_foo() is not None, 'Object collected but contextmanager is active'
    gc.collect()
    assert weak_foo() is None, 'Object not collected but contextmanager has exited'


def verify_concurrent_order(executed, *expected):
    look_at_index = 0
    for expected_group in expected:
        executed_group = set(executed[look_at_index:look_at_index + len(expected_group)])
        assert executed_group == expected_group, 'wrong execution order'
        look_at_index += len(executed_group)
    assert look_at_index == len(executed), 'executed list is shorted than expected'


def test_synchronization_coordinator_wait_for_everyone():
    mo = MultiObject(range(3))

    sync = SynchronizationCoordinator(len(mo))
    executed = []

    def foo(i):
        def execute(caption):
            executed.append((i, caption))

        sleep(i / 10)
        execute('after sleep')
        sync.wait_for_everyone()
        execute('after wait')
        sync.wait_for_everyone()

        sleep(i / 10)
        execute('after sleep 2')
        sync.wait_for_everyone()
        execute('after wait 2')

    mo.call(foo)
    verify_concurrent_order(
        executed,
        {(i, 'after sleep') for i in range(3)},
        {(i, 'after wait') for i in range(3)},
        {(i, 'after sleep 2') for i in range(3)},
        {(i, 'after wait 2') for i in range(3)})


def test_synchronization_coordinator_collect_and_call_once():
    mo = MultiObject(range(3))

    sync = SynchronizationCoordinator(len(mo))
    executed = []

    def foo(i):
        def execute(caption):
            executed.append((i, caption))

        sleep(i / 10)

        def func_to_call_once(param):
            executed.append('params = %s' % sorted(param))
            return sum(param)
        result = sync.collect_and_call_once(i + 1, func_to_call_once)
        execute('result is %s' % result)

        assert sync.collect_and_call_once(i, len) == 3, 'parameters remain from previous call'

    mo.call(foo)
    verify_concurrent_order(
        executed,
        {'params = [1, 2, 3]'},
        {(i, 'result is 6') for i in range(3)})


def test_synchronization_coordinator_abandon():
    mo = MultiObject(range(3))

    sync = SynchronizationCoordinator(len(mo))
    executed = []

    def foo(i):
        def execute(caption):
            executed.append((i, caption))

        sync.wait_for_everyone()
        execute('after wait 1')

        if i == 2:
            sync.abandon()
            return
        # Only two waiters should reach here
        sync.wait_for_everyone()
        execute('after wait 2')

        # Even without explicit call to abandon, sync should only wait for two waiters
        sync.wait_for_everyone()
        execute('after wait 3')

    mo.call(foo)
    verify_concurrent_order(
        executed,
        {(i, 'after wait 1') for i in range(3)},
        {(i, 'after wait 2') for i in range(2)},
        {(i, 'after wait 3') for i in range(2)})


def test_synchronization_coordinator_exception_in_collect_and_call_once():
    mo = MultiObject(range(3))

    sync = SynchronizationCoordinator(len(mo))
    times_called = 0

    class MyException(Exception):
        pass

    def foo(i):
        def func_to_call_once(_):
            nonlocal times_called
            times_called += 1
            raise MyException

        with pytest.raises(MyException):
            sync.collect_and_call_once(i, func_to_call_once)

        assert sync.collect_and_call_once(i + 1, sum) == 6

    mo.call(foo)
    assert times_called == 1, 'collect_and_call_once with exception called the function more than once'


def test_synchronization_coordinator_with_multiobject():
    mo = MultiObject(range(3))

    executed = []

    def foo(i, _sync=SYNC):
        def execute(caption):
            executed.append((i, caption))

        sleep(i / 10)
        _sync.wait_for_everyone()
        execute('after wait')

        def func_to_call_once(param):
            executed.append('params = %s' % sorted(param))
            return sum(param)
        result = _sync.collect_and_call_once(i + 1, func_to_call_once)
        execute('result is %s' % result)

    foo(10)
    assert executed == [
        (10, 'after wait'),
        'params = [11]',
        (10, 'result is 11')]
    executed.clear()

    mo.call(foo)
    verify_concurrent_order(
        executed,
        {(i, 'after wait') for i in range(3)},
        {'params = [1, 2, 3]'},
        {(i, 'result is 6') for i in range(3)})


def test_synchronization_coordinator_with_multiobject_exception():
    mo = MultiObject(range(3))

    executed = []

    class MyException(Exception):
        pass

    def foo(i, _sync=SYNC):
        def execute(caption):
            executed.append((i, caption))

        _sync.wait_for_everyone()
        execute('after wait')

        if i == 2:
            raise MyException

        _sync.wait_for_everyone()
        execute('after wait/abandon')

    with pytest.raises(MultiException) as exc:
        mo.call(foo)
    assert exc.value.count == 1
    assert exc.value.common_type is MyException

    verify_concurrent_order(
        executed,
        {(i, 'after wait') for i in range(3)},
        {(i, 'after wait/abandon') for i in range(2)})


def test_synchronization_coordinator_with_multiobject_early_return():
    mo = MultiObject(range(3))

    executed = []

    def foo(i, _sync=SYNC):
        def execute(caption):
            executed.append((i, caption))

        _sync.wait_for_everyone()
        execute('after wait')

        if i == 2:
            return

        _sync.wait_for_everyone()
        execute('after wait/abandon')

    mo.call(foo)

    verify_concurrent_order(
        executed,
        {(i, 'after wait') for i in range(3)},
        {(i, 'after wait/abandon') for i in range(2)})


def test_synchronization_coordinator_with_multiobject_method():
    class Foo:
        def __init__(self, i):
            self.i = i

        def foo(self, _sync=SYNC):
            return (self.i, _sync.collect_and_call_once(self.i, lambda i_values: sorted(i_values)))

    mo = MultiObject(Foo(i) for i in range(3))

    assert mo.foo().T == (
        (0, [0, 1, 2]),
        (1, [0, 1, 2]),
        (2, [0, 1, 2]))


def test_synchronization_coordinator_failing_context_manager():
    class MyException(Exception):
        pass

    @contextmanager
    def foo(should_fail, _sync=SYNC):
        if should_fail:
            raise MyException()
        else:
            yield

    inside_executed = False
    with pytest.raises(MultiException[MyException]):
        with MultiObject([False, True]).call(foo):
            inside_executed = True

    assert not inside_executed, 'CM body executed even though __enter__ failed in one thread'


def test_synchronization_coordinator_timeout():
    mo = MultiObject(range(3))

    def foo(i, _sync=SYNC):
        sleep(i / 10)
        _sync.wait_for_everyone(timeout=0.1)

    with pytest.raises(MultiException) as exc:
        mo.call(foo)
    assert exc.value.count == len(mo)
    assert exc.value.common_type is threading.BrokenBarrierError


def test_synchronization_coordinator_with_context_manager():
    mo = MultiObject(range(3))

    executed = []

    @contextmanager
    def foo(i, _sync=SYNC):
        def execute(caption):
            executed.append((i, caption))

        sleep(i / 10)
        execute('after sleep')
        _sync.wait_for_everyone()
        execute('before yield')
        yield
        _sync.wait_for_everyone()
        execute('after yield')

    with mo.call(foo):
        executed.append('with body')

    verify_concurrent_order(
        executed,
        {(i, 'after sleep') for i in range(3)},
        {(i, 'before yield') for i in range(3)},
        {'with body'},
        {(i, 'after yield') for i in range(3)})


def test_tag_along_thread():
    counter = 0

    def increment_counter():
        nonlocal counter
        counter += 1
        sleep(0.5)

    tag_along_thread = TagAlongThread(increment_counter, 'counter-incrementer')

    MultiObject(range(8)).call(lambda _: tag_along_thread())

    # The first call should get it's own iteration of the TagAlongThread. The
    # rest should wait for it to finish, and then all use the second iteration.
    # We don't have control over the timing inside the threads though, so we
    # allow fewer or more iterations - as long as an iteration did happen(at
    # least 1) and that invocations did stack together(less than 5 should do
    # it)
    assert 1 <= counter < 5


def test_logged_lock():
    lock = LoggedRLock("test", lease_expiration=1, log_interval=.2)

    step1 = threading.Event()
    step2 = threading.Event()

    def do_lock():
        with lock:
            step1.set()
            step2.wait()

    with concurrent(do_lock):
        # wait for thread to hold the lock
        step1.wait()

        # we'll mock the logger so we can ensure it logged
        with patch("easypy.sync._logger") as _logger:

            assert not lock.acquire(timeout=0.5)  # below the lease_expiration

            # the expiration mechanism should kick in
            with pytest.raises(LockLeaseExpired):
                lock.acquire()

        # let other thread finish
        step2.set()

    with lock:
        pass

    assert sum(c == call("%s - waiting...", lock) for c in _logger.debug.call_args_list) > 3


# this might be useful sometimes, but for now it didn't catch a bug
def disable_test_logged_lock_races():
    lease_expiration = 1
    lock = LoggedRLock("test", lease_expiration=lease_expiration, log_interval=.1)
    import logging

    def do_lock(idx):
        sleep(random.random())
        if lock.acquire(timeout=1, blocking=random.random() > 0.5):
            logging.info("%02d: acquired", idx)
            sleep(random.random() * lease_expiration * 0.9)
            lock.release()
            logging.info("%02d: released", idx)
        else:
            logging.info("%02d: timed out", idx)

    with ExitStack() as stack:
        for i in range(30):
            stack.enter_context(concurrent(do_lock, idx=i, loop=True, sleep=0))
        sleep(5)


def test_sync_singleton():

    class S(metaclass=SynchronizedSingleton):
        def __init__(self):
            sleep(1)

    a, b = MultiObject(range(2)).call(lambda _: S())
    assert a is b


def test_logged_condition():
    cond = LoggedCondition('test', log_interval=.1)

    progress = 0
    executed = []

    def wait_for_progress_to(progress_to):
        cond.wait_for(lambda: progress_to <= progress, 'progress to %s', progress_to)
        executed.append(progress_to)

    with concurrent(wait_for_progress_to, 10), concurrent(wait_for_progress_to, 20), concurrent(wait_for_progress_to, 30):
        with patch("easypy.sync._logger") as _logger:
            sleep(0.3)

        assert any(c == call("%s - waiting for progress to %s", cond, 10) for c in _logger.debug.call_args_list)
        assert any(c == call("%s - waiting for progress to %s", cond, 20) for c in _logger.debug.call_args_list)
        assert any(c == call("%s - waiting for progress to %s", cond, 30) for c in _logger.debug.call_args_list)
        assert executed == []

        with patch("easypy.sync._logger") as _logger:
            with cond.notifying_all('setting progress to 10'):
                progress = 10
        assert [c for c in _logger.debug.call_args_list if 'performed' in c[0][0]] == [
            call("%s - performed: setting progress to 10", cond)]

        with patch("easypy.sync._logger") as _logger:
            sleep(0.3)

        assert not any(c == call("%s - waiting for progress to %s", cond, 10) for c in _logger.debug.call_args_list)
        assert any(c == call("%s - waiting for progress to %s", cond, 20) for c in _logger.debug.call_args_list)
        assert any(c == call("%s - waiting for progress to %s", cond, 30) for c in _logger.debug.call_args_list)
        assert executed == [10]

        with patch("easypy.sync._logger") as _logger:
            with cond.notifying_all('setting progress to 30'):
                progress = 30
        assert [c for c in _logger.debug.call_args_list if 'performed' in c[0][0]] == [
            call("%s - performed: setting progress to 30", cond)]

        with patch("easypy.sync._logger") as _logger:
            sleep(0.3)

        assert not any(c == call("%s - waiting for progress to %s", cond, 10) for c in _logger.debug.call_args_list)
        assert not any(c == call("%s - waiting for progress to %s", cond, 20) for c in _logger.debug.call_args_list)
        assert not any(c == call("%s - waiting for progress to %s", cond, 30) for c in _logger.debug.call_args_list)
        assert executed == [10, 20, 30] or executed == [10, 30, 20]

        with patch("easypy.sync._logger") as _logger:
            with pytest.raises(TimeoutException):
                cond.wait_for(lambda: False, 'the impossible', timeout=1)

        assert sum(c == call("%s - waiting for the impossible", cond) for c in _logger.debug.call_args_list) > 3


def test_logged_condition_exception():
    cond = LoggedCondition('test', log_interval=.2)

    should_throw = False

    class TestException(Exception):
        pass

    def waiter():
        if should_throw:
            raise TestException

    with pytest.raises(TestException):
        with concurrent(cond.wait_for, waiter, 'throw'):
            sleep(0.5)
            should_throw = True


def test_logged_condition_waited_for():
    cond = LoggedCondition('test', log_interval=15)
    progress = 0
    executed = []

    def wait_then_set_to(wait_to, set_to):
        nonlocal progress
        with cond.waited_for(lambda: progress == wait_to, 'progress to be %s', wait_to):
            executed.append('%s -> %s' % (progress, set_to))
            progress = set_to

    with concurrent(wait_then_set_to, 10, 20), concurrent(wait_then_set_to, 20, 30), concurrent(wait_then_set_to, 30, 40):
        with cond.notifying_all('setting progress to 10'):
            progress = 10

    assert executed == ['10 -> 20', '20 -> 30', '30 -> 40']


def test_wait_exception():
    with pytest.raises(Exception, match=".*`message` is required.*"):
        wait(0.1, pred=lambda: True)

    wait(0.1)
    wait(0.1, pred=lambda: True, message='message')
    wait(0.1, pred=lambda: True, message=False)
    repeat(0.1, callback=lambda: True)


def test_wait_better_exception():

    class TimedOut(PredicateNotSatisfied):
        pass

    i = 0

    def check():
        nonlocal i
        i += 1
        if i < 3:
            raise TimedOut(a=1, b=2)
        return True

    with pytest.raises(TimedOut):
        # due to the short timeout and long sleep, the pred would called exactly twice
        wait(.1, pred=check, sleep=1, message=False)

    assert i == 2
    wait(.1, pred=check, message=False)


def test_wait_better_exception_nested():

    class TimedOut(PredicateNotSatisfied):
        pass

    i = 0

    def check():
        nonlocal i
        i += 1
        if i < 3:
            raise TimedOut(a=1, b=2)
        return True

    with pytest.raises(TimedOut):
        # due to the short timeout and long sleep, the pred would called exactly twice
        # also, the external wait should call the inner one only once, due to the TimedOut exception,
        # which it knows not to swallow
        wait(5, lambda: wait(.1, pred=check, sleep=1, message=False), sleep=1, message=False)

    assert i == 2
    wait(.1, pred=check, message=False)


def test_iter_wait_warning():
    with pytest.raises(Exception, match=".*`message` is required.*"):
        for _ in iter_wait(0.1, pred=lambda: True):
            pass

    no_warn_iters = [
        iter_wait(0.1),
        iter_wait(0.1, pred=lambda: True, message='message'),
        iter_wait(0.1, pred=lambda: True, throw=False),
        iter_wait(0.1, pred=lambda: True, message=False)
    ]
    for i in no_warn_iters:
        for _ in i:
            pass


def test_iter_wait_progress_inbetween_sleep():
    data = Bunch(a=3)

    def get():
        data.a -= 1
        return data.a

    sleep = .1
    g = iter_wait_progress(get, advance_timeout=10, sleep=sleep)

    # first iteration should be immediate
    t = Timer()
    next(g)
    assert t.duration < sleep

    # subsequent iteration should be at least 'sleep' long
    next(g)
    assert t.duration >= sleep

    for state in g:
        pass
    assert state.finished is True


def test_iter_wait_progress_total_timeout():
    data = Bunch(a=1000)

    def get():
        data.a -= 1
        return data.a

    with pytest.raises(TimeoutException) as exc:
        for state in iter_wait_progress(get, advance_timeout=1, sleep=.05, total_timeout=.1):
            pass
    assert exc.value.message.startswith("advanced but failed to finish")


def test_wait_long_predicate():
    """
    After the actual check the predicate is held for .3 seconds. Make sure
    that we don't get a timeout after .2 seconds - because the actual
    condition should be met in .1 second!
    """

    t = Timer()

    def pred():
        try:
            return 0.1 < t.duration
        finally:
            wait(0.3)

    wait(0.2, pred, message=False)


def test_timeout_exception():
    exc = None

    with timing() as t:
        try:
            wait(0.5, lambda: False, message=False)
        except TimeoutException as e:
            exc = e

    assert exc.duration > 0.5
    assert exc.start_time >= t.start_time
    assert exc.start_time < t.stop_time
    assert t.duration > exc.duration


def test_wait_with_callable_message():
    val = ['FOO']

    with pytest.raises(TimeoutException) as e:
        def pred():
            val[0] = 'BAR'
            return False
        wait(pred=pred, timeout=.1, message=lambda: 'val is %s' % val[0])

    assert val[0] == 'BAR'
    assert e.value.message == 'val is BAR'


@pytest.mark.parametrize("multipred", [False, True])
def test_wait_do_something_on_final_attempt(multipred):
    data = []

    def pred(is_final_attempt):
        if is_final_attempt:
            data.append('final')
        data.append('regular')
        return False

    if multipred:
        pred = [pred]

    with pytest.raises(TimeoutException):
        wait(pred=pred, timeout=.5, sleep=.1, message=False)

    assert all(iteration == 'regular' for iteration in data[:-2])
    assert data[-2] == 'final'
    assert data[-1] == 'regular'


def test_wait_log_predicate(get_log):
    def pred():
        raise PredicateNotSatisfied('bad attempt')

    with pytest.raises(TimeoutException):
        wait(pred=pred, timeout=.5, sleep=.1, message=False, log_interval=0.2)
    durations = re.findall('Still waiting after (.*?): bad attempt', get_log())
    rounded_durations = [round(Duration(d), 2) for d in durations]
    assert rounded_durations == [0.2, 0.4], 'expected logs at 200ms and 400ms, got %s' % (durations,)



from io import StringIO

from easypy.collections import defaultlist
from easypy.colors import colorize, uncolored
from easypy.humanize import compact


class Column():
    def __init__(self, name, title=None, max_width=None, align=None, header_align=None, padding=None, drop_if_empty=False):
        self.name = name
        self.max_width = max_width
        self.align = align
        self.header_align = header_align
        self.padding = padding
        self.overflow = 'ellipsis'
        self.title = title or name
        self.drop_if_empty = drop_if_empty
        self.visible = True


class Table():
    """
    :param List[Column] columns: column descriptors
    :param List[Bunch] data: rows
    """

    HEADER_SEP = "|"
    SEPARATORS = "|"
    BAR = '='
    BAR_SEP = ":"

    def __init__(self, *columns, data=None, max_col_width=None, align='left', header_align='center', padding=1):
        self.data = data or []
        self.columns = []

        self.max_col_width = max_col_width
        self.align = align
        self.header_align = header_align
        self.padding = padding

        for column in columns:
            self.add_column(column)

    _ALIGN_MAP = dict(left='<', right='>', center='^')

    def add_column(self, column: Column):
        self.columns.append(column)

    def add_row(self, **row):
        self.data.append(row)

    def render(self):
        rendered = defaultlist(list)
        columns = []

        def _get_value(data, value):
            ret = data.get(value)
            if ret is None:
                ret = ''
            return ret

        for column in self.columns:
            if not column.visible:
                continue
            rows = [_get_value(data, column.name) for data in self.data]
            if not any(filter(lambda i: i != '', rows)) and column.drop_if_empty:
                continue
            columns.append(column)

            if column.max_width is None:
                column.max_width = self.max_col_width
            if column.align is None:
                column.align = self.align
            if column.header_align is None:
                column.header_align = self.header_align
            if column.padding is None:
                column.padding = self.padding

            raw_data = [column.title] + rows
            colored_data = [colorize(str(data)) for data in raw_data]
            uncolored_data = [uncolored(data) for data in colored_data]
            max_width = column.max_width or max(len(data) for data in uncolored_data)
            for i, data in enumerate(colored_data):
                align = column.header_align if i == 0 else column.align
                coloring_spacing = len(colored_data[i]) - len(uncolored_data[i])
                spacing = max_width + coloring_spacing
                format_string = "{{data:{align}{spacing}}}".format(align=self._ALIGN_MAP[align], spacing=spacing)
                rendered[i].append(format_string.format(data=data))

        output = StringIO()
        for r_i, row in enumerate(rendered):
            r_parts = []

            sep = self.HEADER_SEP if r_i == 0 else self.SEPARATORS[r_i % len(self.SEPARATORS)]

            for col_i, col in enumerate(row):
                column = columns[col_i]
                padding = column.padding * " "
                if column.max_width and r_i > 0:
                    col = compact(col, column.max_width, suffix_length=column.max_width // 10)
                r_parts.append("{padding}{col}{padding}".format(col=col, padding=padding))

            output.write(sep.join(r_parts))
            output.write("\n")

            if r_i == 0:
                r_parts = [self.BAR * len(uncolored(part)) for part in r_parts]
                output.write(self.BAR_SEP.join(r_parts))
                output.write("\n")

        output.seek(0)
        return output.read()


class DecoratedTable(Table):
    HEADER_SEP = "│"
    SEPARATORS = "┼│┊┊│"
    BAR = '═'
    BAR_SEP = "╪"


def _test():
    table = Table(Column("first", "GREEN<<First>>"))
    table.add_column(Column("second", align='right'))
    table.add_row(first='1', second='BLUE<<longer>> second MAGENTA<<column>>')
    table.add_row(first='longer first column', second='2')
    print(table.render())



from __future__ import absolute_import

from functools import partial, wraps
from contextlib import contextmanager
from typing import Union
from easypy.timing import Timer
from easypy.units import Duration
import random
import time

from .decorations import parametrizeable_decorator


import logging
from logging import getLogger
_logger = getLogger(__name__)


UNACCEPTABLE_EXCEPTIONS = (NameError, AttributeError, TypeError, KeyboardInterrupt)


class ExponentialBackoff:
    """
    Factory for increasing backoffs.

    :param initial: The value to return on the first call.
    :param base: The ratio between two consecutive returned values.
    :param maximum: A limit to the value increasing.

    Each call generates a backoff (number) that's greater than the previous::

        >>> backoff = ExponentialBackoff(initial=1, base=2)
        >>> backoff()
        1
        >>> backoff()
        2
        >>> backoff()
        4
        >>> backoff()
        8
        >>> backoff()
        16
    """

    def __init__(self, initial=1, maximum=30, base=1.5):
        self.base = base
        self.initial = initial
        self.current = initial
        self.maximum = float(maximum)

    def get_current(self):
        """Get the current value without progressing."""
        return self.current

    def __call__(self):
        """Get the current value and progress to the next one."""
        ret = min(self.get_current(), self.maximum)
        self.current = min(ret * self.base, self.maximum)
        return ret

    def __repr__(self):
        return "{0.__class__.__name__}({0.base}, {0.initial}, {0.maximum} {0.current})".format(self)


class RandomExponentialBackoff(ExponentialBackoff):
    """Variation of ``ExponentialBackoff`` that adds randomization to the returned values."""

    def get_current(self):
        return random.random() * self.current + self.initial


class ExpiringCounter(object):
    """
    A counter that decreases every time its expiration is checked.

    >>> counter = ExpiringCounter(3)
    >>> counter.expired
    False
    >>> counter.expired
    False
    >>> counter.expired
    False
    >>> counter.expired
    True
    """
    def __init__(self, times):
        self.times = times

    @property
    def expired(self):
        """Progress the counter and return whether or not it expired."""
        self.times -= 1
        return self.times < 0

    @property
    def remain(self):
        """Return the number of remaining times to check ``expired`` before it expires."""
        return self.times


def retry(times: Union[int, Duration, Timer], func, args=[], kwargs={}, acceptable=Exception, sleep=1,
          max_sleep=False, log_level=logging.DEBUG, pred=None, unacceptable=()):
    """
    Runs a function again and again until it finishes without throwing.

    :param times: Limit number of attempts before errors are propagated instead of suppressed.
                  If an `int`, the execution is retried at most `times` times.
                  If a `Timer`, retries until the `times` expires.
                  If a `Duration`, retires for the specified duarion.
    :param func: The function to run.
    :param list args: Positional arguments for the function.
    :param dict kwargs: Keyword arguments for the function.
    :param acceptable: Exception (or tuple of exceptions) to repeat the function on.
    :param sleep: Time to wait between multiple retries.
    :param max_sleep: If set, increase the sleep after each retry until reaching this number.
    :param log_level: Before retrying, log the exceptions at that level.
    :param pred: If set, decide whether or not to retry on an exception.
    :param unacceptable: Exception (or tuple of exceptions) to not repeat the function on and re-raise them instead.
    """

    if unacceptable is None:
        unacceptable = ()
    elif isinstance(unacceptable, tuple):
        unacceptable += UNACCEPTABLE_EXCEPTIONS
    else:
        unacceptable = (unacceptable,) + UNACCEPTABLE_EXCEPTIONS

    if isinstance(times, Timer):
        stopper = times  # a timer is a valid stopper
    elif isinstance(times, Duration):
        stopper = Timer(expiration=times)
    elif isinstance(times, int):
        stopper = ExpiringCounter(times)
    else:
        assert False, "'times' must be an 'int', 'Duration' or 'Timer', got %r" % times

    if max_sleep:
        sleep = RandomExponentialBackoff(sleep, max_sleep)
    if not pred:
        pred = lambda exc: True

    while True:
        try:
            return func(*args, **kwargs)
        except unacceptable as exc:
            raise
        except acceptable as exc:
            raise_if_async_exception(exc)
            if not pred(exc):
                raise
            if stopper.expired:
                raise
            _logger.log(log_level, "Exception thrown: %r", exc)
            sleep_for = 0
            if sleep:
                # support for ExponentialBackoff
                sleep_for = sleep() if callable(sleep) else sleep
            _logger.log(log_level, "Retrying... (%r remain) in %s seconds", stopper.remain, sleep_for)
            time.sleep(sleep_for)


def retrying(times: Union[int, Duration, Timer], acceptable=Exception, sleep=1, max_sleep=False, log_level=logging.DEBUG, pred=None, unacceptable=()):
    """Try running the decorated function, retrying if an acceptable exception caught.

    times - limit number of attempts before errors are propagated instead of suppressed.
            if an `int`, the execution is retried at most `times` times.
            if a `Timer`, retries until the `times` expires.
            if a `Duration`, retries up until `Timer(times)` expires
    acceptable - exception or tuple of exceptions which to catch and retry upon.
    sleep - time to wait between attempts. can be a callable.
    max_sleep - if given, then the time to sleep between attempts is `RandomExponentialBackoff(initial=sleep, maximum=max_sleep)`.
    log_level - level of the log to emit when catching an exception and retrying.
    pred - if given, then retries only if pred(exception) is True.

    >>> @retrying(times=5)
    ... def get_lucky():
    ...     if random.random() < 0.5:
    ...         raise Exception('No luck')
    ...     return 'Got lucky'
    """
    def wrapper(func):
        @wraps(func)
        def impl(*args, **kwargs):
            return retry(
                times, func, args, kwargs, sleep=sleep, max_sleep=max_sleep, acceptable=acceptable,
                log_level=log_level, pred=pred, unacceptable=unacceptable)
        return impl
    return wrapper


class _Retry(Exception):
    pass


retrying.Retry = _Retry

retrying.debug = partial(retrying, log_level=logging.DEBUG)
retrying.info = partial(retrying, log_level=logging.INFO)
retrying.warning = partial(retrying, log_level=logging.WARNING)
retrying.error = partial(retrying, log_level=logging.ERROR)


@parametrizeable_decorator
def resilient(func=None, default=None, **kw):
    """Suppress exceptions raised from the decorated function.

    default - value to return if an exception is suppressed.
    kw      - see `resilience`

    >>> @resilient(default=0)
    ... def get_number():
    ...     return int(open('non-existent-file').read())
    >>> get_number()
    0
    """
    kw.setdefault('msg', "ignoring error in %s ({type})" % func.__qualname__)

    @wraps(func)
    def inner(*args, **kwargs):
        with resilience(**kw):
            return func(*args, **kwargs)
        return default  # we reach here only if an exception was caught and handled by resilience
    return inner


@contextmanager
def resilience(msg="ignoring error {type}", acceptable=Exception, unacceptable=(), log_level=logging.DEBUG, pred=None):
    """Suppress exceptions raised from the wrapped scope.

    msg          - format of log to print when an exception is suppressed.
    acceptable   - exception or tuple of exceptions which to suppress.
    unacceptable - exception or tuple of exception which to not suppress, even if they are in `acceptable`.
                   the exceptions in UNACCEPTABLE_EXCEPTIONS are always unacceptable, unless `unacceptable` is None.
    log_level    - level of the log to emit when suppressing an exception.
    pred         - if given, then an exception is suppressed only if pred(exception) is True.

    >>> import errno
    >>> with resilience(acceptable=OSError, pred=lambda ex: ex.errno == errno.ENOENT):
    ...     print('before')
    ...     open('non-existent-file')
    ...     print('after')
    before
    """
    if unacceptable is None:
        unacceptable = ()
    elif isinstance(unacceptable, tuple):
        unacceptable += UNACCEPTABLE_EXCEPTIONS
    else:
        unacceptable = (unacceptable,) + UNACCEPTABLE_EXCEPTIONS
    try:
        yield
    except unacceptable as exc:
        raise
    except acceptable as exc:
        if pred and not pred(exc):
            raise
        raise_if_async_exception(exc)
        _logger.log(log_level, msg.format(exc=exc, type=exc.__class__.__qualname__))
        if log_level > logging.DEBUG:
            _logger.debug("Traceback:", exc_info=True)


resilient.debug = partial(resilient, log_level=logging.DEBUG)
resilient.info = partial(resilient, log_level=logging.INFO)
resilient.warning = partial(resilient, log_level=logging.WARNING)
resilient.error = partial(resilient, log_level=logging.ERROR)

resilience.debug = partial(resilience, log_level=logging.DEBUG)
resilience.info = partial(resilience, log_level=logging.INFO)
resilience.warning = partial(resilience, log_level=logging.WARNING)
resilience.error = partial(resilience, log_level=logging.ERROR)


def raise_if_async_exception(exc):
    """
    Helper for handling exceptions potentially invoked in another thread.

    This is so that exception raised from other threads don't get supressed by retry/resilience.
    See :meth:`~easypy.concurrency.raise_in_main_thread`.
    """
    if getattr(exc, "_raised_asynchronously", False):
        _logger.info('Raising asynchronous error')
        raise exc



from contextlib import contextmanager
import time
from datetime import datetime, timedelta

from .decorations import parametrizeable_decorator
from .units import Duration


inf = float("inf")  # same as math.inf, but compatible with 3.4


def to_timestamp(t):
    return t.timestamp() if isinstance(t, datetime) else t


class TimeInterval(object):
    __slots__ = ("t0", "t1")

    NO_START = -inf
    NO_END = inf

    def __init__(self, from_time=NO_START, to_time=NO_END):
        self.t0 = self.NO_START if from_time is None else to_timestamp(from_time)
        self.t1 = self.NO_END if to_time is None else to_timestamp(to_time)

    @property
    def start_time(self):
        if self.t0 > self.NO_START:
            return datetime.fromtimestamp(self.t0)

    @property
    def end_time(self):
        if self.t1 < self.NO_END:
            return datetime.fromtimestamp(self.t1)

    @property
    def duration_delta(self):
        return timedelta(seconds=self.duration) if self.duration < inf else None

    @property
    def duration(self):
        return Duration(self.t1 - self.t0)

    def __contains__(self, t):
        """
        check if timestamp is within the timer's interval
        """

        if isinstance(t, Timer):
            t = t.to_interval()

        if isinstance(t, TimeInterval):
            return self.t0 <= t.t0 and t.t1 <= self.t1

        t = to_timestamp(t)
        return self.t0 <= t <= self.t1

    def to_timer(self):
        assert self.t0 > self.NO_START, "Can't convert a start-less TimeInterval to a Timer"
        t = Timer(self.t0)
        t.t1 = self.t1 if self.t1 < self.NO_END else None
        return t

    def render(self):
        t0, duration, t1 = self.t0, self.duration, self.t1
        if (t0, t1) == (self.NO_START, self.NO_END):
            return "Eternity"

        parts = [".", "."]

        if t0 > self.NO_START:
            parts[0] = time.strftime("%T", time.localtime(t0))

        if t1 < self.NO_END:
            parts[-1] = time.strftime("%T", time.localtime(self.t1))

        if duration < inf:
            parts.insert(1, "({})".format(duration))

        return "..".join(parts)

    def __str__(self):
        return "<TI %s>" % self.render()

    def __repr__(self):
        return "TimeInterval(%s)" % self.render()

    def _repr_pretty_(self, p, cycle):
        # used by IPython
        from easypy.colors import MAGENTA
        if cycle:
            p.text('TimeInterval(...)')
            return
        p.text(MAGENTA(self.render()))


class Timer(object):

    """
    Multi-purpose timer object::

        t = Timer()
        # do something ...
        t.stop()
        print(t.elapsed)

        t = Timer(expiration=120)
        while not t.expired:
            print("Time remaining: %r" % t.remain)
            # do something ...

        for remain in Timer(expiration=120).iter(sleep=2):
            print("Time remaining: %r" % remain)
            # do something ...
    """

    __slots__ = ("t0", "t1", "expiration")

    def __init__(self, now=None, expiration=None):
        self.reset(now)
        self.t1 = None
        self.expiration = None if expiration is None else Duration(expiration)

    def reset(self, now=None):
        self.t0 = now or time.time()

    def stop(self):
        self.t1 = time.time()
        return self.t1 - self.t0

    def iter(self, sleep=1):
        while not self.expired:
            time.sleep(sleep)
            yield self.remain

    def to_interval(self):
        return TimeInterval(self.t0, self.t1)

    def __contains__(self, timestamp):
        """
        check if timestamp is within the timer's interval
        """
        timestamp = to_timestamp(timestamp)
        if timestamp < self.t0:
            return False
        if not self.t1:
            return True
        if timestamp > self.t1:
            return False
        return True

    __iter__ = iter

    @property
    def stopped(self):
        return self.t1 is not None

    @property
    def duration_delta(self):
        return timedelta(seconds=self.duration)

    @property
    def duration(self):
        return Duration((self.t1 or time.time()) - self.t0)

    @property
    def elapsed_delta(self):
        return timedelta(seconds=time.time() - self.t0)

    @property
    def elapsed(self):
        return Duration(self.elapsed_delta.total_seconds())

    @property
    def expired(self):
        return Duration(max(0, self.elapsed - self.expiration) if self.expiration is not None else 0)

    @property
    def remain(self):
        return Duration(max(0, self.expiration - self.elapsed)) if self.expiration is not None else None

    @property
    def start_time(self):
        return datetime.fromtimestamp(self.t0)

    @property
    def stop_time(self):
        if self.t1 is None:
            return None
        else:
            return datetime.fromtimestamp(self.t1)

    def render(self):
        t0, duration, elapsed, expired, stopped = self.t0, self.duration, self.elapsed, self.expired, self.stopped
        st = time.strftime("%T", time.localtime(t0))
        if expired:
            duration = "{}+{}".format(self.expiration, expired)
        if stopped:
            et = time.strftime("%T", time.localtime(self.t1))
            fmt = "{st}..({duration})..{et}"
        elif expired:
            fmt = "{st}..({duration})"
        else:
            fmt = "{st}..({duration})"
        return fmt.format(**locals())

    def __str__(self):
        return "<T %s>" % self.render()

    def __repr__(self):
        return "Timer(%s)" % self.render()

    def _repr_pretty_(self, p, cycle):
        # used by IPython
        from easypy.colors import MAGENTA, RED
        if cycle:
            p.text('Timer(...)')
            return
        p.text((RED if self.expired else MAGENTA)(self.render()))


class BackoffTimer(Timer):

    def __init__(self, expiration, now=None, backoff_every=5, backoff_by=5, max_interval=None):
        super().__init__(expiration=expiration, now=now)
        self.backoff_by = backoff_by
        self.backoff_every = backoff_every
        self.iteration = 0
        self.max_interval = max_interval

    def backoff(self):
        self.iteration += 1
        if self.iteration % self.backoff_every == 0:
            self.expiration *= self.backoff_by
            if self.max_interval:
                self.expiration = min(self.max_interval, self.expiration)
        super().reset()


class StopWatch(object):
    def __init__(self):
        self._last_pause = 0
        self._timer = Timer()
        self.paused = True

    @property
    def elapsed(self):
        if self.paused:
            return self._last_pause
        return self._last_pause + self._timer.elapsed

    def pause(self):
        self._last_pause += self._timer.elapsed
        self.paused = True

    def start(self):
        self.paused = False
        self._timer.reset()

    def reset_and_start(self, last_paused=None):
        self._last_pause = 0
        self.start()


@contextmanager
def timing(t=None):
    if not t:
        t = Timer()
    try:
        yield t
    finally:
        t.stop()


def repeat(timeout, callback, sleep=0.5, progressbar=True):
    pred = lambda: callback() and False  # prevent 'wait' from stopping when the callback returns a nonzero
    return wait(timeout, pred=pred, sleep=sleep, progressbar=progressbar, throw=False)


@parametrizeable_decorator
def at_period(func, period=None):
    if not period:
        raise NotImplementedError('must specify a period')
    timer = Timer(expiration=period)

    def inner(*args, **kwargs):
        if timer.expired:
            ret = func(*args, **kwargs)
            timer.reset()
            return ret
    return inner


class StateTimeHistogram(object):
    def __init__(self):
        self._state = None
        self._states_times = {}
        self._timer = Timer()

    def _update_state_time(self, states_times=None):
        if states_times is None:
            states_times = self._states_times
        if self._state is not None:
            states_times[self._state] = states_times.get(self._state, 0) + self._timer.elapsed

    def finish(self):
        self._update_state_time()
        self._state = None
        self._timer.stop()

    def set_state(self, state):
        self._update_state_time()
        self._state = state
        self._timer.reset()

    @property
    def states_times(self):
        states_times = dict(self._states_times)
        self._update_state_time(states_times)
        return states_times


def throttled(duration):
    """
    Syntax sugar over timecache decorator.

    With accent on throttling calls and not actual caching of values Concurrent
    callers will block if function is executing, since they might depend on
    side effect of function call.
    """
    from easypy.caching import timecache
    return timecache(expiration=duration)


# re-exports
from .sync import wait, PredicateNotSatisfied, TimeoutException, iter_wait, wait_progress, iter_wait_progress



import inspect
import weakref
from functools import wraps, update_wrapper
from types import MethodType
from .decorations import parametrizeable_decorator
from .collections import ilistify, intersected_dict
from .tokens import Token  # for backwards compatibility


class Hex(int):

    def __str__(self):
        return "%X" % self

    def __repr__(self):
        return "0x%x" % self


def get_all_subclasses(cls, include_mixins=False):

    def is_mixin(subclass):
        return getattr(subclass, "_%s__is_mixin" % subclass.__name__, False)

    def gen(cls):
        for subclass in cls.__subclasses__():
            if include_mixins or not is_mixin(subclass):
                yield subclass
            yield from gen(subclass)

    return list(gen(cls))


def stack_level_to_get_out_of_file():
    frame = inspect.currentframe().f_back
    filename = frame.f_code.co_filename
    stack_levels = 1
    while frame.f_code.co_filename == filename:
        stack_levels += 1
        frame = frame.f_back
    return stack_levels


def at_most(val, mx_val):
    return min(val, mx_val)


def at_least(val, mn_val):
    return max(val, mn_val)


def clamp(val, at_least, at_most):
    """
    Clamps a value so it doesn't exceed specified limits.
    If one of the edges is not needed, it should be passed as None (consider using at_most / at_least functions).
    :param at_least: Minimum possible value.
    :param at_most: Maxium possible value.
    :return: The clamped value.
    """

    if at_least > at_most:
        raise ValueError("Min value cannot be higher than max value.")

    if at_most is not None:
        val = min(at_most, val)
    if at_least is not None:
        val = max(at_least, val)
    return val


class WeakMethodDead(Exception):
    pass


class WeakMethodWrapper:
    def __init__(self, weak_method):
        if isinstance(weak_method, MethodType):
            weak_method = weakref.WeakMethod(weak_method)
        self.weak_method = weak_method
        update_wrapper(self, weak_method(), updated=())
        self.__wrapped__ = weak_method

    def __call__(self, *args, **kwargs):
        method = self.weak_method()
        if method is None:
            raise WeakMethodDead
        return method(*args, **kwargs)


@parametrizeable_decorator
def kwargs_resilient(func, negligible=None):
    """
    If function does not specify **kwargs, pass only params which it can accept

    :param negligible: If set, only be resilient to these specific parameters:

                - Other parameters will be passed normally, even if they don't appear in the signature.
                - If a specified parameter is not in the signature, don't pass it even if there are **kwargs.
    """
    if isinstance(func, weakref.WeakMethod):
        spec = inspect.getfullargspec(inspect.unwrap(func()))
        func = WeakMethodWrapper(func)
    else:
        spec = inspect.getfullargspec(inspect.unwrap(func))
    acceptable_args = set(spec.args or ())
    if isinstance(func, MethodType):
        acceptable_args -= {spec.args[0]}

    if negligible is None:
        @wraps(func)
        def inner(*args, **kwargs):
            if spec.varkw is None:
                kwargs = intersected_dict(kwargs, acceptable_args)
            return func(*args, **kwargs)
    else:
        negligible = set(ilistify(negligible))

        @wraps(func)
        def inner(*args, **kwargs):
            kwargs = {k: v for k, v in kwargs.items()
                      if k in acceptable_args
                      or k not in negligible}
            return func(*args, **kwargs)

    return inner



from __future__ import absolute_import
import inspect
import threading
import functools
from itertools import chain, count
import logging
from enum import Enum
from contextlib import contextmanager, ExitStack
import weakref

from easypy.concurrency import Futures, MultiObject
from easypy.decorations import parametrizeable_decorator
from easypy.exceptions import TException
from easypy.contexts import is_contextmanager
from easypy.misc import kwargs_resilient, WeakMethodDead
from easypy.collections import separate

PRIORITIES = Enum("PRIORITIES", "FIRST NONE LAST")
_logger = logging.getLogger(__name__)


ids = set()


class MissingIdentifier(TException):
    template = "signal handler {_signal_name} must be called with a '{_identifier}' argument"


def make_id(name):
    # from md5 import md5
    # id = md5(name).hexdigest()[:5].upper()
    for c in range(10):
        id = "".join(word[0] for word in name.split("_")) + (str(c) if c else "")
        if id not in ids:
            break
    ids.add(id)
    return id


def get_original_func(func):
    while True:
        if isinstance(func, functools.partial):
            func = func.func
        elif hasattr(func, "__wrapped__") and not (
            is_contextmanager(func) and not is_contextmanager(func.__wrapped__)
            # This means func is the context manager itself (which wraps a non-context-manager function)
        ):
            func = func.__wrapped__
        else:
            return func


class RemoveHandler(Exception):
    pass


STALE_HANDLER = (WeakMethodDead, RemoveHandler)


class SignalHandler(object):

    _idx_gen = count(100)

    def __init__(self, func, asynchronous=False, priority=PRIORITIES.NONE, times=None, identifier=None, **kw):
        self.func = kwargs_resilient(func)
        self._func = get_original_func(func)  # save the original funtion so we can unregister based on the function

        # identifier only applicable to methods
        if inspect.ismethod(self._func) or isinstance(self._func, weakref.WeakMethod):
            self.identifier = identifier
        else:
            self.identifier = None

        # Backwards compatibility
        self.asynchronous = kw.pop("async", asynchronous)
        assert not kw, "SignalHandler's kwargs should contain only 'async' argument for backwards compatibility"

        self.priority = priority
        self.times = times
        self.idx = next(self._idx_gen)

        if isinstance(func, weakref.WeakMethod):
            func = func()  # to allow accessing it's __code__ and __name__
        self.filename = func.__code__.co_filename
        self.lineno = func.__code__.co_firstlineno
        self.name = self.__name__ = func.__name__

    def __repr__(self):
        return "<handler #{0.idx} '{0.name}' ({0.filename}:{0.lineno})>".format(self)

    def should_run(self, **kwargs):
        if self.times == 0:
            return False

        if self.identifier:
            try:
                handler_object = self._func.__self__
            except AttributeError:
                func = self._func()
                if func is None:
                    raise WeakMethodDead
                handler_object = func.__self__
            target_object = kwargs[self.identifier]

            if hasattr(self._func, 'identifier_path'):
                # special case if the signal method defined a user-defined attribute path for fiding the associated identifier
                path = self._func.identifier_path
                handler_object = eval('obj.%s' % path.strip('.'), dict(obj=handler_object), {})
            elif hasattr(handler_object, self.identifier):
                # the identifier exists as an attribute on the handler object, named the same
                handler_object = getattr(handler_object, self.identifier)
            elif isinstance(handler_object, target_object.__class__):
                # the handler_object is itself of the same type as the target_object, so we should regard it is the identifier itself
                pass
            else:
                # the handler_object isn't identifiable, so we'll just pass it the identifier and hope it knows what to do
                handler_object = None

            if handler_object is not None and handler_object != target_object:
                return False
        return True

    def __call__(self, *, swallow_exceptions, **kwargs):
        if self.times is not None:
            self.times -= 1

        try:
            return self.func(**kwargs)
        except STALE_HANDLER:
            self.times = 0
            raise
        except:
            if not swallow_exceptions:
                raise
            _logger.silent_exception("Exception in (%s) ignored", self)

    @property
    def bound_object(self):
        if isinstance(self._func, weakref.WeakMethod):
            return weakref.ref.__call__(self._func)
        else:
            return None


class Signal:

    ALL = {}

    def __new__(cls, name, asynchronous=None, swallow_exceptions=False, log=True):
        try:
            return cls.ALL[name]
        except KeyError:
            pass
        assert threading.main_thread() is threading.current_thread(), "Can only create Signal objects from the MainThread"
        signal = object.__new__(cls)
        # we use this to track the calling of various callbacks. we use this short id so it doesn't overflow the logs
        signal.id = make_id(name)
        signal.handlers = {priority: [] for priority in PRIORITIES}
        signal.name = name
        signal.swallow_exceptions = swallow_exceptions
        signal.asynchronous = asynchronous
        signal.identifier = None
        signal.log = log
        return cls.ALL.setdefault(name, signal)

    def iter_handlers(self):
        return chain(*(self.handlers[k] for k in PRIORITIES))

    def iter_handlers_by_priority(self):
        for k in PRIORITIES:
            yield separate(self.handlers[k], lambda h: h.asynchronous)

    def remove_handler(self, handler):
        self.handlers[handler.priority].remove(handler)
        _logger.debug("handler removed from '%s' (%s): %s", self.name, handler.priority.name, handler)

    def remove_handlers_if_exist(self, handlers):
        if not handlers:
            return
        handlers = set(handlers)
        for priority in {h.priority for h in handlers}:
            self.handlers[priority][:] = (handler for handler in self.handlers[priority] if handler not in handlers)

    def register(self, func=None, asynchronous=None, priority=PRIORITIES.NONE, times=None, **kw):
        # backwards compatibility
        asynchronous = kw.pop('async', asynchronous)
        assert not kw

        if not func:
            return functools.partial(self.register, asynchronous=asynchronous, priority=priority)
        if asynchronous is None:
            asynchronous = False if self.asynchronous is None else self.asynchronous
        elif self.asynchronous is not None:
            assert self.asynchronous == asynchronous, "Signal is set with asynchronous=%s" % self.asynchronous
        handler = SignalHandler(func, asynchronous, priority, times=times, identifier=self.identifier)
        self.handlers[priority].append(handler)
        _logger.debug("registered handler for '%s' (%s): %s", self.name, priority.name, handler)
        return func

    def unregister(self, func):
        for handler in self.iter_handlers():
            if func in (
                    handler._func,  # simple case
                    getattr(handler._func, "__wrapped__", None),  # wrapped with @wraps
                    getattr(handler._func, "func", None),  # wrapped with 'partial'
                    handler._func() if isinstance(handler._func, weakref.WeakMethod) else None,
                    ):
                self.remove_handler(handler)

    @contextmanager
    def registered(self, func, **kwargs):
        self.register(func, **kwargs)
        try:
            yield
        finally:
            self.unregister(func)

    def __call__(self, **kwargs):
        if not self.identifier:
            pass
        elif self.identifier in kwargs:
            pass
        else:
            raise MissingIdentifier(_signal_name=self.name, _identifier=self.identifier)

        if self.log:
            _logger.debug("Triggered %s", self)

        for handler in self.iter_handlers():
            if handler.times == 0:
                self.remove_handler(handler)

        kwargs.setdefault('swallow_exceptions', self.swallow_exceptions)

        with ExitStack() as STACK:
            STACK.enter_context(_logger.context(self.id))

            if any(h.asynchronous for h in self.iter_handlers()):
                futures = STACK.enter_context(Futures.executor())
            else:
                futures = None

            handlers_to_remove = []

            def run_handler(handler):
                try:
                    with _logger.context("#%03d" % handler.idx):
                        if not handler.should_run(**kwargs):
                            return
                        handler(**kwargs)
                except STALE_HANDLER:
                    handlers_to_remove.append(handler)

            for async_handlers, synced_handlers in self.iter_handlers_by_priority():

                for handler in async_handlers:
                    futures.submit(run_handler, handler)

                for handler in synced_handlers:
                    run_handler(handler)

                if futures:
                    for future in futures.as_completed():
                        future.result()        # bubble up exceptions

        self.remove_handlers_if_exist(handlers_to_remove)

    def __str__(self):
        return "<Signal %s (%s)>" % (self.name, self.id)

    __repr__ = __str__


@parametrizeable_decorator
def signal_identifier_path(func, path):
    func.identifier_path = path
    return func


class ContextManagerSignal(Signal):
    @contextmanager
    def __call__(self, **kwargs):
        if self.log:
            # log signal for centralized logging analytics.
            # minimize text message as most of the data is sent in the 'extra' dict
            # signal fields get prefixed with 'sig_', and the values are repr'ed
            signal_fields = {("sig_%s" % k): repr(v) for (k, v) in kwargs.items()}
            _logger.debug("Triggered '%s' (%s) - entering", self.name, self.id,
                          extra=dict(signal_fields, signal=self.name, signal_id=self.id))

        for handler in self.iter_handlers():
            if handler.times == 0:
                self.unregister(handler.func)

        kwargs.setdefault('swallow_exceptions', self.swallow_exceptions)

        with ExitStack() as handlers_stack:
            handlers_to_remove = []

            def should_run(handler):
                try:
                    return handler.should_run(**kwargs)
                except STALE_HANDLER:
                    pass
                handlers_to_remove.append(handler)
                return False

            indexer = count()

            @contextmanager
            def run_handler(handler):
                index = next(indexer)
                already_yielded = False
                try:
                    with _logger.context(self.id), _logger.context("%02d" % index):
                        with handler(**kwargs):
                            yield
                            already_yielded = True
                except STALE_HANDLER:
                    handlers_to_remove.append(handler)
                    if not already_yielded:
                        yield

            for async_handlers, synced_handlers in self.iter_handlers_by_priority():
                async_handlers = MultiObject(filter(should_run, async_handlers))
                synced_handlers = list(filter(should_run, synced_handlers))

                if async_handlers:
                    handlers_stack.enter_context(MultiObject(async_handlers).call(run_handler))

                for handler in synced_handlers:
                    handlers_stack.enter_context(run_handler(handler))
            yield

        self.remove_handlers_if_exist(handlers_to_remove)


####################################
def register_signal(name, func, **kwargs):
    signal = Signal(name)
    signal.register(func, **kwargs)
    return functools.partial(signal.unregister, func)


def unregister_signal(name, func):
    try:
        signal = Signal.ALL[name]
    except KeyError:
        return
    return signal.unregister(func)


@parametrizeable_decorator
def register(func, **kwargs):
    register_signal(func.__name__, func, **kwargs)
    return func


def unregister(func):
    return unregister_signal(func.__name__, func)


def _set_handler_params(**kw):
    def inner(func):
        params = getattr(func, '_signal_handler_params', {})
        params.update(kw)
        func._signal_handler_params = params
        return func
    return inner


run_first = _set_handler_params(priority=PRIORITIES.FIRST)
run_last = _set_handler_params(priority=PRIORITIES.LAST)
run_async = _set_handler_params(asynchronous=True)
run_sync = _set_handler_params(asynchronous=False)


@functools.lru_cache(None)
def get_signals_for_type(typ):
    return {n for typ in inspect.getmro(typ)
            for n in dir(typ) if n.startswith("on_")}


class __FakeCode:
    def __init__(self, fake_filename, fake_firstlineno):
        self.co_filename = fake_filename
        self.co_firstlineno = fake_firstlineno


def __fake_meth_type(func, obj, name, code):
    method = func.__get__(obj, type(obj))
    method.__name__ = name
    method.__code__ = code
    return method


def register_object(obj, **kwargs):
    for method_name in get_signals_for_type(type(obj)):
        method = getattr(obj, method_name)
        assert callable(method), "'%s' (%r) is not callable" % (method, obj)

        # Don't use static/class methods for automatic event registration - they
        # could be registered multiple times if the are multiple objects of that
        # type in the system, or not at all if there are no objects of that type
        # in the system.
        assert method is not getattr(type(obj), method_name, None), "'%s' is a static/class method" % method

        params = getattr(method, '_signal_handler_params', {})
        intersection = set(params).intersection(kwargs)
        assert not intersection, "parameter conflict in signal object registration (%s)" % (intersection)
        params.update(kwargs)

        method_name, *_ = method_name.partition("__")  # allows multiple methods for the same signal

        if not inspect.ismethod(method):
            # Fake method attributes for WeakMethod
            method.__self__ = obj
            method.__func__ = getattr(type(obj), method_name)
            if not hasattr(method, '__code__'):
                method.__code__ = __FakeCode('', '')
            fake_meth_type = functools.partial(
                __fake_meth_type,
                # Best effort: fake name and code for SignalHandler. We don't know the line number.
                name=method_name,
                code=__FakeCode(type(obj).__module__, '?'))

        method = weakref.WeakMethod(method)
        try:
            method._meth_type = fake_meth_type
        except UnboundLocalError:
            pass

        register_signal(method_name, method, **params)


def unregister_object(obj):
    for signal_name in {method_name.split('__', 1)[0] for method_name in get_signals_for_type(type(obj))}:
        for handlers in Signal.ALL[signal_name].handlers.values():
            handlers[:] = (handler for handler in handlers if handler.bound_object is not obj)


def call_signal(name, **kwargs):
    return Signal(name)(**kwargs)


def log_all_signal_ids(logger=_logger, level=logging.DEBUG):
    for signal in sorted(Signal.ALL.itervalues(), key=lambda signal: signal.name):
        logger.log(level, "%s - '%s'", signal.id, signal.name)


# ===================================================================================================
# Module hack: ``from easypy.signals import on_some_signal``
#              ``from easypy.signals import on_ctx_some_signal`` for a context-manager signal
# ===================================================================================================
def _hack():
    import sys
    from types import ModuleType
    this = sys.modules[__name__]

    class SignalsModule(ModuleType):
        """The module-hack that allows us to use ``from easypy.signals import on_some_signal``"""
        __all__ = ()  # to make help() happy
        __package__ = __name__
        __path__ = []
        __file__ = __file__

        def __getattr__(self, name):
            try:
                return getattr(this, name)
            except AttributeError:
                if name.startswith("on_ctx_"):
                    return ContextManagerSignal(name)
                elif name.startswith("on_"):
                    return Signal(name)
                raise

        def __dir__(self):
            return sorted(dir(this) + list(Signal.ALL))

    module = SignalsModule(__name__, SignalsModule.__doc__)
    sys.modules[module.__name__] = module

_hack()



import pytest

from contextlib import contextmanager

from easypy.signals import register_object, unregister_object, MissingIdentifier, PRIORITIES
from easypy.signals import on_test, on_async_test
from easypy.signals import on_test_identifier
from easypy.signals import on_ctx_test
from easypy.signals import on_ctx_test_identifier

on_test_identifier.identifier = 'obj'
on_ctx_test_identifier.identifier = 'obj'
on_async_test.asynchronous = True


def test_simple_signal_registration():

    @on_test.register
    def foo(a, b):
        a / b

    with pytest.raises(ZeroDivisionError):
        on_test(a=5, b=0, c='c')

    on_test.unregister(foo)
    on_test(a=5, b=0, c='c')


def test_simple_signal_object_registration():

    class Foo():
        def on_test(self, a, b):
            a / b
        __init__ = register_object

    f = Foo()

    with pytest.raises(ZeroDivisionError):
        on_test(a=5, b=0, c='c')

    unregister_object(f)


def test_simple_signal_object_identifier():

    on_test_identifier(obj='xxx')

    class Foo():
        def on_test_identifier(self, a):
            a / self.b
        __init__ = register_object

    f1 = Foo()
    f2 = Foo()

    f1.b = 1
    f2.b = 0

    with pytest.raises(MissingIdentifier):
        on_test_identifier(a=5, b=0, c='c')

    on_test_identifier(a=5, b=0, obj=f1)

    with pytest.raises(ZeroDivisionError):
        on_test_identifier(a=5, b=0, obj=f2)

    unregister_object(f1)
    unregister_object(f2)

    with on_test_identifier.registered(f2.on_test_identifier):
        with pytest.raises(ZeroDivisionError):
            on_test_identifier(a=5, b=0, obj=f2)


def test_simple_signal_object_wo_identifier():

    class Foo():
        def on_test_identifier(self, obj):
            1 / 0
        __init__ = register_object

    f1 = Foo()

    with pytest.raises(ZeroDivisionError):
        on_test_identifier(a=5, b=0, obj='obj1')
        on_test_identifier(a=5, b=0, obj='obj2')

    unregister_object(f1)


def test_simple_signal_object_identifier_attribute():

    class Foo():
        def on_test_identifier(self):
            1 / 0
        __init__ = register_object

    f1 = Foo()
    f1.obj = 'obj1'

    on_test_identifier(obj='xxx')

    with pytest.raises(ZeroDivisionError):
        on_test_identifier(obj='obj1')

    unregister_object(f1)


def test_registration_context():

    def foo(a, b):
        a / b

    with on_test.registered(foo):
        with pytest.raises(ZeroDivisionError):
            on_test(a=5, b=0, c='c')

    on_test(a=5, b=0, c='c')


def test_priorities():

    l = []

    @on_test.register(priority=PRIORITIES.LAST)
    def last():
        l.append(3)

    @on_test.register(priority=PRIORITIES.FIRST)
    def first():
        l.append(1)

    @on_test.register()
    def mid():
        l.append(2)

    on_test()

    assert l == [1, 2, 3]


def test_priorities_async():

    l = []
    import time

    @on_test.register(asynchronous=True, priority=PRIORITIES.FIRST)
    def first():
        time.sleep(.05)
        l.append(1)

    @on_test.register(asynchronous=True, priority=PRIORITIES.LAST)
    def last():
        l.append(2)

    on_test()

    assert l == [1, 2]


def test_async():
    from threading import get_ident

    main = get_ident()

    @on_async_test.register(asynchronous=True)
    def a1():
        assert get_ident() != main

    @on_async_test.register()  # follows the current setting
    def a2():
        assert get_ident() != main

    with pytest.raises(AssertionError):
        @on_async_test.register(asynchronous=False)
        def xx():
            assert get_ident() == main

    on_async_test()


def test_ctx():
    result = []

    class Foo:
        @contextmanager
        def on_ctx_test(self, before, after):
            result.append(before)
            yield
            result.append(after)

    foo = Foo()

    register_object(foo)

    assert result == []
    with on_ctx_test(before=1, after=2):
        assert result == [1]
    assert result == [1, 2]

    unregister_object(foo)

    with on_ctx_test(before=3, after=4):
        assert result == [1, 2]
    assert result == [1, 2]


def test_ctx_with_identifier():
    class Foo():
        @contextmanager
        def on_ctx_test_identifier(self, before, after):
            self.result.append(before)
            yield
            self.result.append(after)

    f1 = Foo()
    f1.obj = 'obj1'
    f1.result = []
    f2 = Foo()
    f2.obj = 'obj2'
    f2.result = []

    register_object(f1)
    register_object(f2)

    with on_ctx_test_identifier(before=1, after=2, obj='obj1'):
        assert f1.result == [1]
        assert f2.result == []
    assert f1.result == [1, 2]
    assert f2.result == []

    with on_ctx_test_identifier(before=3, after=4, obj='obj2'):
        assert f1.result == [1, 2]
        assert f2.result == [3]
    assert f1.result == [1, 2]
    assert f2.result == [3, 4]


def test_signal_weakref():
    """
    Test that signals handlers of methods are deleted when their objects get collected
    """
    import gc

    class Foo:
        def on_test(self, a, b):
            a / b

    foo = Foo()
    register_object(foo)

    with pytest.raises(ZeroDivisionError):
        on_test(a=5, b=0, c='c')

    del foo
    gc.collect()

    on_test(a=5, b=0, c='c')


def test_signal_weakref_complex_descriptors():
    import gc
    from easypy.lockstep import lockstep

    class Foo:
        @lockstep
        def on_test(self, a, b):
            a / b

    foo = Foo()
    register_object(foo)

    with pytest.raises(ZeroDivisionError):
        on_test(a=5, b=0, c='c')

    del foo
    gc.collect()

    on_test(a=5, b=0, c='c')


def test_signal_weakref_context_manager_delete_after():
    import gc

    result = []

    class Foo:
        @contextmanager
        def on_ctx_test(self, before, after):
            result.append(before)
            yield
            result.append(after)

    foo = Foo()
    register_object(foo)

    with on_ctx_test(before=1, after=2):
        assert result == [1]
    assert result == [1, 2]

    del foo
    gc.collect()

    with on_ctx_test(before=3, after=4):
        assert result == [1, 2]
    assert result == [1, 2]


def test_signal_weakref_context_manager_delete_during():
    import gc

    result = []

    class Foo:
        @contextmanager
        def on_ctx_test(self, before, after):
            result.append(before)
            yield
            result.append(after)

    foo = Foo()
    register_object(foo)

    with on_ctx_test(before=1, after=2):
        assert result == [1]
    assert result == [1, 2]

    with on_ctx_test(before=3, after=4):
        assert result == [1, 2, 3]
        del foo
        gc.collect()
    # The context manager should keep the signal handler alive
    assert result == [1, 2, 3, 4]


def test_signal_weakref_with_identifier():
    import gc

    class Foo:
        def on_test_identifier(self, a, b):
            a / b

    foo = Foo()
    foo.obj = 'obj'
    register_object(foo)

    with pytest.raises(ZeroDivisionError):
        on_test_identifier(a=5, b=0, obj='obj')
    on_test_identifier(a=5, b=0, obj='noobj')

    del foo
    gc.collect()

    on_test_identifier(a=5, b=0, obj='obj')
    on_test_identifier(a=5, b=0, obj='noobj')


def test_signal_weakref_context_manager_delete_after_with_identifier():
    import gc

    result = []

    class Foo:
        @contextmanager
        def on_ctx_test_identifier(self, before, after):
            result.append(before)
            yield
            result.append(after)

    foo = Foo()
    foo.obj = 'obj'
    register_object(foo)

    foo2 = Foo()
    foo2.obj = 'otherobj'
    register_object(foo2)

    with on_ctx_test_identifier(before=1, after=2, obj='obj'):
        assert result == [1]
    assert result == [1, 2]

    del foo
    del foo2
    gc.collect()

    with on_ctx_test_identifier(before=3, after=4, obj='obj'):
        assert result == [1, 2]
    assert result == [1, 2]


def test_signal_weakref_context_manager_delete_during_with_identifier():
    import gc

    result = []

    class Foo:
        @contextmanager
        def on_ctx_test_identifier(self, before, after):
            result.append(before)
            yield
            result.append(after)

    foo = Foo()
    foo.obj = 'obj'
    register_object(foo)

    foo2 = Foo()
    foo2.obj = 'otherobj'
    register_object(foo2)

    with on_ctx_test_identifier(before=1, after=2, obj='obj'):
        assert result == [1]
    assert result == [1, 2]

    with on_ctx_test_identifier(before=3, after=4, obj='obj'):
        assert result == [1, 2, 3]
        del foo
        del foo2
        gc.collect()
    # The context manager should keep the signal handler alive
    assert result == [1, 2, 3, 4]



from easypy.exceptions import TException
from easypy.bunch import Bunch


class T(TException):
    template = "The happened: {what}"


def test_pickle_texception():
    import pickle

    t1 = T(what="happened", a=1, b=Bunch(x=[1, 2, 3], y=range(5)))
    t2 = pickle.loads(pickle.dumps(t1))

    assert t1.render() == t2.render()
    assert t1._params == t2._params



import os
if os.getenv("GEVENT") == "true":
    from easypy.gevent import apply_patch
    apply_patch()


import logging
import logging.config
from easypy import logging as easypy_logging
logging.addLevelName(logging.WARN, "WARN")  # instead of "WARNING", so that it takes less space...
logging.addLevelName(logging.NOTSET, "NA")  # instead of "NOTSET", so that it takes less space...
if not issubclass(logging.Logger, easypy_logging.ContextLoggerMixin):
    logging.Logger.__bases__ = logging.Logger.__bases__ + (easypy_logging.ContextLoggerMixin,)


logging.basicConfig(level=logging.DEBUG, format='%(asctime)s|%(process)2s:%(threadName)-25s|%(name)-40s|%(levelname)-5s|%(funcName)-30s |%(message)s')



import inspect
from abc import ABCMeta
from functools import wraps
from collections import OrderedDict

from .misc import kwargs_resilient
from .collections import as_list


class EasyMeta(ABCMeta):
    """
    This class helps in implementing various metaclass-based magic.
    Implement the ``before_subclass_init`` and/or ``after_subclass_init`` (see defined in :class:``EasyMetaHooks``)
    to modify the class spec, register subclasses, etc.

    Each hook method can be defined more than once (with the same name), and methods will all be invoked sequentially.

    Important: the hooks are not invoked on the class that implements the hooks - only on it subclasses.
    """

    class EasyMetaHooks:
        """
        The class defines the available EasyMeta hooks (slots), and registers handlers for EasyMeta derivatives
        """

        HOOK_NAMES = []

        def _define_hook(dlg, HOOK_NAMES=HOOK_NAMES):
            HOOK_NAMES.append(dlg.__name__)

            @wraps(dlg)
            def hook(self, *args, **kwargs):
                kwargs_resilience = kwargs_resilient(negligible=self.class_kwargs.keys())
                kwargs.update((k, v) for k, v in self.class_kwargs.items() if k not in kwargs)

                for hook in self._em_hooks[dlg.__name__]:
                    kwargs_resilience(hook)(*args, **kwargs)

            return hook

        @_define_hook
        def after_subclass_init(self, cls):
            """
            Invoked after a subclass is being initialized

            >>> class PrintTheName(metaclass=EasyMeta):
            ...     @EasyMeta.Hook
            ...     def after_subclass_init(cls):
            ...         print('Declared', cls.__name__)
            ...
            ...
            >>> class Foo(PrintTheName):
            ...     pass
            Declared Foo
            """

        @_define_hook
        def before_subclass_init(self, name, bases, dct):
            """
            Invoked after a subclass is being initialized

            >>> class AddMember(metaclass=EasyMeta):
            ...     @EasyMeta.Hook
            ...     def before_subclass_init(name, bases, dct):
            ...         dct['foo'] = 'bar'
            ...
            ...
            >>> class Foo(AddMember):
            ...     pass
            >>> Foo.foo
            'bar'
            """

        def __init__(self, class_kwargs={}):
            self._em_hooks = {name: [] for name in self.HOOK_NAMES}
            self.class_kwargs = class_kwargs

        def add(self, hook):
            self._em_hooks[hook.__name__].append(hook)

        def extend(self, other):
            for k, v in other._em_hooks.items():
                slot = self._em_hooks[k]
                slot.extend(hook for hook in v if hook not in slot)

    class EasyMetaDslDict(OrderedDict):
        """
        This class is used as the namespace for the user's class.
        Any member decorated as an EasyMeta hook gets removed from the namespace and
        put into a special 'hooks' collection. The EasyMeta metaclass then invokes the
        delegates registered under those hooks
        """
        def __init__(self, class_name):
            super().__init__()
            self._class_name = class_name
            self._em_hooks = EasyMeta.EasyMetaHooks()

        def __setitem__(self, name, value):
            if isinstance(value, EasyMeta.Hook):
                self._em_hooks.add(value.dlg)
            else:
                return super().__setitem__(name, value)

    @classmethod
    def __prepare__(metacls, name, bases, **kwds):
        dsl = metacls.EasyMetaDslDict(class_name=name)
        return dsl

    class Hook(object):
        def __init__(self, dlg):
            self.dlg = dlg

    def __init__(cls, name, bases, dct, **kwargs):
        super().__init__(name, bases, dct)

    def __new__(mcs, name, bases, dct, **kwargs):
        aggregated_hooks = mcs.EasyMetaHooks(class_kwargs=kwargs)

        bases = list(bases)  # allow the hook to modify the base class list

        for base in bases:
            for sub_base in reversed(inspect.getmro(base)):
                if isinstance(sub_base, EasyMeta):
                    aggregated_hooks.extend(sub_base._em_hooks)

        ns = dict(dct)
        aggregated_hooks.before_subclass_init(name, bases, ns)
        new_type = super().__new__(mcs, name, tuple(bases), ns)
        aggregated_hooks.after_subclass_init(new_type)

        new_type._em_hooks = dct._em_hooks

        return new_type


class GetAllSubclasses(metaclass=EasyMeta):
    """
    Meta-magic mixin for registering subclasses

    The ``get_all_subclasses`` class method will return a list of all subclasses
    of the class it was called on. The class it was called on is not included in
    the list.

    >>> class Foo(GetAllSubclasses):
    ...     pass
    ...
    ...
    >>> class Bar(Foo):
    ...     pass
    ...
    ...
    >>> class Baz(Foo):
    ...     pass
    ...
    ...
    >>> class Qux(Bar):
    ...     pass
    ...
    ...
    >>> Foo.get_all_subclasses()
    [<class 'easypy.meta.Bar'>, <class 'easypy.meta.Qux'>, <class 'easypy.meta.Baz'>]
    >>> Bar.get_all_subclasses()
    [<class 'easypy.meta.Qux'>]
    >>> Baz.get_all_subclasses()
    []
    >>> Qux.get_all_subclasses()
    []
    """

    @EasyMeta.Hook
    def after_subclass_init(cls):
        cls.__direct_subclasses = []
        for base in cls.__bases__:
            if base is not GetAllSubclasses and issubclass(base, GetAllSubclasses):
                base.__direct_subclasses.append(cls)

    @classmethod
    @as_list
    def get_subclasses(cls):
        """
        List immediate subclasses of this class
        """
        yield from cls.__direct_subclasses

    @classmethod
    def iter_all_subclasses(cls, level=0) -> (int, type):
        """
        walk all subclasses of this class
        """
        for subclass in cls.__direct_subclasses:
            yield level, subclass
            yield from subclass.iter_all_subclasses(level=level + 1)

    @classmethod
    @as_list
    def get_all_subclasses(cls):
        """
        List all subclasses of this class
        """
        for level, subclass in cls.iter_all_subclasses():
            yield subclass



from easypy.units import byte, KiB


def test_data_sizes():
    assert "{0!s}, {0!r}, {0:10text}".format(byte) == "byte, byte,     1 byte"
    assert "{0!s}, {0!r}, {0:10text}".format(1000 * byte) == "KB, KB,       1 KB"
    assert "{0!s}, {0!r}, {0:10text}".format(1020 * byte) == "1020bytes, 1020*bytes,  1020bytes"
    assert "{0!s}, {0!r}, {0:10text}".format(1024 * byte) == "KiB, KiB,      1 KiB"
    assert "{0!s}, {0!r}, {0:10text}".format(2**20 * byte) == "MiB, MiB,      1 MiB"
    assert "{0!s}, {0!r}, {0:10text}".format(2**21 * byte) == "2MiB, 2*MiB,       2MiB"
    assert "{0!s}, {0!r}, {0:10text}".format(2**21 * byte + 100) == "~2.0MiB, 2097252*bytes,    ~2.0MiB"
    assert "{0!s}, {0!r}, {0:10text}".format(2**41 * byte + 100) == "~2.0TiB, 2199023255652*bytes,    ~2.0TiB"


def test_operators():

    assert (byte * 1024) == KiB
    assert KiB / 1024 == byte
    assert KiB / KiB == 1
    assert KiB / 7 == 146.28571428571428
    assert KiB // 7 == 146
    assert 2050 // KiB == (2 * byte)

    # check that __r*__ overloads are used when the unit doesn't support the right-hand operand
    class Foo():
        def __rfloordiv__(self, div):
            return self

    foo = Foo()

    assert KiB // foo is foo



from easypy.semver import SemVer, SMV


def test_loads():
    semver = SemVer.loads('3.4.5.6-hello')
    assert semver.build == 6
    assert semver.patch == 5
    assert semver.minor == 4
    assert semver.major == 3
    assert semver.tag == 'hello'

    semver = SemVer.loads('3_4_5:hello', separator='_', tag_separator=':')
    assert semver.build is None
    assert semver.patch == 5
    assert semver.minor == 4
    assert semver.major == 3
    assert semver.tag == 'hello'


def test_dumps():
    version = '3.4.5.6-hello'
    assert SemVer.loads(version).dumps() == version

    string = SemVer.loads(version).dumps(separator='_', tag_separator=':')
    assert string == version.replace('.', '_').replace('-', ':')


def test_copy():
    semver = SemVer.loads('1000.2000.3000.5000-10102010')
    assert semver.copy() == semver

    assert semver.copy(major=1500).major == 1500


def test_comparison():
    semver1 = SemVer.loads('2.1-aaa')
    semver2 = SemVer.loads('2.1-bbb')
    semver3 = SemVer.loads('2.1.2-a')
    semver4 = SemVer.loads('2.2')
    assert semver1 < semver2 < semver3 < semver4

    assert semver1 == semver2.copy(tag='aaa')


def test_bumping():
    semver = SemVer.loads('1.1.1.1-aaa')
    assert semver.bump_build().dumps() == '1.1.1.2'

    # Build part is only printed when it is set
    assert semver.bump_patch().dumps() == '1.1.2'
    assert semver.bump_minor().dumps() == '1.2.0'
    assert semver.bump_major().dumps() == '2.0.0'
    assert semver.bump_major(clear_tag=False).dumps() == '2.0.0-aaa'



"""
This module is about making it easier to create decorators
"""

from functools import wraps, partial, update_wrapper
from operator import attrgetter
from abc import ABCMeta, abstractmethod
import inspect

from easypy.exceptions import TException


def parametrizeable_decorator(deco):
    @wraps(deco)
    def inner(func=None, **kwargs):
        if func is None:
            return partial(deco, **kwargs)
        else:
            return wraps(func)(deco(func, **kwargs))
    return inner


def wrapper_decorator(deco):
    @wraps(deco)
    def inner(func):
        return wraps(func)(deco(func))
    return inner


def reusable_contextmanager(context_manager):
    """
    Allows the generator-based context manager to be used more than once
    """

    if not hasattr(context_manager, '_recreate_cm'):
        return context_manager  # context manager is already reusable (was not created usin yield funcion

    class ReusableCtx:
        def __enter__(self):
            self.cm = context_manager._recreate_cm()
            return self.cm.__enter__()

        def __exit__(self, *args):
            self.cm.__exit__(*args)

    return ReusableCtx()


class DecoratingDescriptor(metaclass=ABCMeta):
    """
    Base class for descriptors that decorate a function

    :param func: The function to be decorated.
    :param bool cached: If ``True``, the decoration will only be done once per instance.

    Use this as a base class for other descriptors. When used on class objects,
    this will return itself. When used on instances, it this will call ``_decorate``
    on the method created by binding ``func``.
    """

    def __init__(self, *, func, cached: bool):
        self._func = func
        self._cached = cached
        self.__property_name = '__property_%s' % id(self)
        update_wrapper(self, func, updated=())

    @abstractmethod
    def _decorate(self, method, instance, owner):
        """
        Override to perform the actual decoration.

        :param method: The method from binding ``func``.
        :params instance: The binding instance (same as in ``__get__``)
        :params owner: The owner class (same as in ``__get__``)
        """
        pass

    def __get__(self, instance, owner):
        method = self._func.__get__(instance, owner)
        if instance is None:
            return method
        else:
            if self._cached:
                try:
                    return instance.__dict__[self.__property_name]
                except KeyError:
                    bound = self._decorate(method, instance, owner)
                    instance.__dict__[self.__property_name] = bound
                    return bound
            else:
                return self._decorate(method, instance, owner)


class LazyDecoratorDescriptor(DecoratingDescriptor):
    def __init__(self, decorator_factory, func, cached):
        super().__init__(func=func, cached=cached)
        self.decorator_factory = decorator_factory

    def _decorate(self, method, instance, owner):
        decorator = self.decorator_factory(instance)
        return decorator(method)


def lazy_decorator(decorator_factory, cached=False):
    """
    Create and apply a decorator only after the method is instantiated::

    :param decorator_factory: A function that will be called with the ``self`` argument.
                              Should return a decorator for the method.
                              If ``string``, use an attribute of ``self`` with that name
                              as the decorator.
    :param bool cached: If ``True``, the decoration will only be done once per instance.

        class UsageWithLambda:
            @lazy_decorator(lambda self: some_decorator_that_needs_the_object(self))
            def foo(self):
                # ...

        class UsageWithAttribute:
            def decorator_method(self, func):
                # ...

            @lazy_decorator('decorator_method')
            def foo(self):
                # ...

        class UsageCached:
            # Without ``cached=True``, this will create a new ``timecache`` on every invocation.
            @lazy_decorator(lambda self: timecache(expiration=1, get_ts_func=lambda: self.ts), cached=True)
            def foo(self):
                # ...
    """

    if callable(decorator_factory):
        pass
    elif isinstance(decorator_factory, str):
        decorator_factory = attrgetter(decorator_factory)
    else:
        raise TypeError('decorator_factory must be callable or string, not %s' % type(decorator_factory))

    def wrapper(func):
        return LazyDecoratorDescriptor(decorator_factory, func, cached)
    return wrapper


class DefaultsMismatch(TException):
    template = 'The defaults of {func} differ from those of {source_of_truth} in params {param_names}'


def ensure_same_defaults(source_of_truth, ignore=()):
    """
    Ensure the decorated function has the same default as the source of truth in optional parameters shared by both

    :param source_of_truth: A function to check the defaults against.
    :param ignore: A list of parameters to ignore even if they exist and have defaults in both functions.
    :raises DefaultsMismatch: When the defaults are different.

    >>> def foo(a=1, b=2, c=3):
    ...     ...
    >>> @ensure_same_defaults(foo)
    ... def bar(a=1, b=2, c=3):  # these defaults are verified by the decorator
    ...         ...
    """

    sot_signature = inspect.signature(source_of_truth)
    params_with_defaults = [
        param for param in sot_signature.parameters.values()
        if param.default is not param.empty
        and param.name not in ignore]

    def gen_mismatches(func):
        signature = inspect.signature(func)
        for sot_param in params_with_defaults:
            param = signature.parameters.get(sot_param.name)
            if param is None:
                continue
            if param.default is param.empty:
                continue
            if sot_param.default != param.default:
                yield sot_param.name

    def wrapper(func):
        mismatches = list(gen_mismatches(func))
        if mismatches:
            raise DefaultsMismatch(
                func=func,
                source_of_truth=source_of_truth,
                param_names=mismatches)
        return func
    return wrapper



import pytest
from easypy.contexts import contextmanager, breakable_section


X = []


@contextmanager
def ctx():
    X.append(1)
    yield 2
    X.pop(0)


def test_simple():
    with ctx() as i:
        assert i == 2
        assert X == [1]
    assert not X


def test_function():
    @ctx()
    def foo():
        assert X == [1]
    foo()
    assert not X


def test_generator():
    @ctx()
    def foo():
        yield from range(5)

    for i in foo():
        assert X == [1]
    assert not X


def test_ctx():
    @ctx()
    @contextmanager
    def foo():
        yield

    with foo():
        assert X == [1]

    assert not X


def test_breakable_section():

    a = []
    with breakable_section() as Break1:
        with breakable_section() as Break2:
            with breakable_section() as Break3:
                raise Break2()
                a += [1]  # this will be skipped
            a += [2]  # this will be skipped
        a += [3]  # landing here
    a += [4]

    assert Break1 is not Break2
    assert Break2 is not Break3
    assert Break3 is not Break1
    assert a == [3, 4]



import re
from itertools import repeat
import random
import numbers
from easypy.exceptions import TException


class Percentage(float):

    def __repr__(self):
        return "{:.1f}%".format(self * 100)

    def __format__(self, spec):
        width, precision, name = re.match(r"(?:(\d+))?(?:\.(\d*))?(\w+)?", spec).groups()
        precision = int(precision) if precision else 1
        if name in ("f", None):
            ret = "{:.{precision}f}%".format(self * 100, precision=precision)
        elif name == "d":
            ret = "{}%".format(round(self * 100))
        return "{:>{width}}".format(ret, width=width or "")

    def _repr_pretty_(self, p, cycle):
        # used by IPython
        from easypy.colors import MAGENTA
        from easypy.humanize import vertbar
        if cycle:
            p.text('Percentage(...)')
            return
        p.text(MAGENTA("{1}:{0!r}".format(self, vertbar(self))))


class UnknownDataSizeError(TException):
    template = 'Could not parse size {}'


class DataSize(int):
    BINARY_UNITS = {unit_name: 1024 ** i for i, unit_name in enumerate(['byte', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'XiB'])}
    DECIMAL_UNITS = {unit_name: 1000 ** i for i, unit_name in enumerate(['KB', 'MB', 'GB', 'TB', 'PB', 'XB'], 1)}
    NAMED_UNITS = dict(BINARY_UNITS, **DECIMAL_UNITS)
    UNIT_NAMES = {unit: name for name, unit in NAMED_UNITS.items()}
    SORTED_UNITS = sorted(BINARY_UNITS.values(), reverse=True)

    def __new__(cls, value):
        if (isinstance(value, str)):
            i, u = re.match(r"(\d*(?:\.\d+)?)?(\w*)", value).groups()
            i = 1 if not i else float(i) if "." in i else int(i)
            value = i * (cls.NAMED_UNITS[u] if u else 1)
            if (isinstance(value, cls)):
                return value

        return super(DataSize, cls).__new__(cls, value)

    def __abs__(self):
        return self.__class__(int.__abs__(self))

    def __add__(self, other):
        return self.__class__(int.__add__(self, other))
    __radd__ = __add__

    def __sub__(self, other):
        return self.__class__(int.__sub__(self, other))

    def __rsub__(self, other):
        return (-self) + other

    def __mul__(self, mul):
        if isinstance(mul, self.__class__):
            raise TypeError('Cannot multiply %s by %s' % (self, mul))
        if not isinstance(mul, numbers.Real):
            raise TypeError('Cannot multiply %s by %r' % (self, mul))
        return self.__class__(int(self) * mul)
    __rmul__ = __mul__

    def __floordiv__(self, div):
        if not isinstance(div, numbers.Number):
            return NotImplemented
        res = int.__floordiv__(self, div)
        if not isinstance(div, self.__class__):
            # if div is not a datasize, return a datasize
            return self.__class__(res)
        # otherwise return an int
        return res

    def __rfloordiv__(self, div):
        if not isinstance(div, numbers.Number):
            return NotImplemented
        res = int.__rfloordiv__(self, div)
        if not isinstance(div, self.__class__):
            # if div is not a datasize, return a datasize
            return self.__class__(res)
        # otherwise return an int
        return res

    def __mod__(self, div):
        return self.__class__(int.__mod__(self, div))

    def __rmod__(self, number):
        if number != 0:
            raise ArithmeticError("Cannot perform modulu of %s by %s" % (number, self))
        return 0

    def __neg__(self):
        return -1*self

    def rounddown(self, unit):
        return (self // unit) * unit

    def roundup(self, unit):
        if 0 == self % unit:
            return self
        else:
            return (self // unit + 1) * unit

    def round(self, unit):
        if self % unit > unit / 2:
            add = unit
        else:
            add = 0
        return self.rounddown(unit) + add

    def ceildiv(self, other):
        return self.roundup(other) / other

    def randrange(self, divisor, start=None):
        'returns a random size below self that divides by divisor'
        if start is None:
            start = self.__class__(0)
        start = start.ceildiv(divisor)
        stop = self.ceildiv(divisor)
        return random.randrange(start=start, stop=stop) * divisor

    def __format__(self, spec):
        width, precision, name = re.match(r"(?:(\d+))?(?:\.(\d*))?(\w+)?", spec).groups()
        if name == "d":
            ret = "{:d}".format(int(self))
        elif name == "text":
            ret = self.render(humanize=True)
        elif name == "f":
            ret = "{:.{precision}f}".format(float(self), precision=precision or 1)
        elif name in ("b", "byte", "bytes"):
            ret = "{:d}*bytes".format(int(self))
        elif name:
            ret = "{:.{precision}f}*{unit}".format(self / self.NAMED_UNITS[name], unit=name, precision=precision or 1)
        else:
            ret = repr(self)
        return "{:>{width}}".format(ret, width=width or "")

    def render(self, humanize=True):
        if 0 == self:
            return '0 bytes' if humanize else '0'

        if self in self.UNIT_NAMES:
            unit = self.UNIT_NAMES[self]
            return '1 {}'.format(unit) if humanize else unit

        for unit in self.SORTED_UNITS:
            name = self.UNIT_NAMES[unit]
            many = 'bytes' if name == 'byte' else name
            if self % unit == 0:
                return '%d%s' % (self / unit, many)
            if self >= unit or (self >= unit / 10 and self * 10 % unit == 0):
                return '~%.1f%s' % (self / unit, many)

        assert False, "This Should not happen"

    def __str__(self):
        return self.render(humanize=False)

    def __repr__(self):
        if self == 0:
            return '0'
        if self in self.UNIT_NAMES:
            return self.UNIT_NAMES[self]
        for unit in self.SORTED_UNITS:
            name = self.UNIT_NAMES[unit]
            many = 'bytes' if name == 'byte' else name
            if self % unit == 0:
                return '%d*%s' % (self / unit, many)
        assert False, "This Should not happen"

    def _repr_pretty_(self, p, cycle):
        # used by IPython
        from easypy.colors import BLUE
        if cycle:
            p.text('DataSize(...)')
            return
        p.text(BLUE(self))


class Duration(float):
    # NAMED_UNITS = reduce(lambda a, b: dict(a, _last=b[0], **{b[0]: a[a['_last']]*int(b[1])}),
    #     (p.split(":") for p in "s:1000 m:60 h:60 d:24".split()),
    #     dict(ms=1, _last='ms'))
    NAMED_UNITS = dict(ms=1/1000, s=1, m=60, h=60*60, d=24*60*60)
    UNIT_NAMES = {unit: name for name, unit in NAMED_UNITS.items()}
    SORTED_UNITS = sorted(NAMED_UNITS.values(), reverse=True)

    def __new__(cls, value):
        if (isinstance(value, str)):
            try:
                i, u = float(value), None
            except ValueError:
                i, u = re.match(r"(\d*(?:\.\d+)?)?(\w*)", value).groups()
                i = 1.0 if not i else float(i)
            value = i * (cls.NAMED_UNITS[u] if u else 1)
            if (isinstance(value, cls)):
                return value

        return super(Duration, cls).__new__(cls, value)

    def __abs__(self):
        return self.__class__(float.__abs__(self))

    def __add__(self, other):
        return self.__class__(float.__add__(self, other))
    __radd__ = __add__

    def __sub__(self, other):
        return self.__class__(float.__sub__(self, other))

    def __rsub__(self, other):
        return (-self) + other

    def __mul__(self, mul):
        if isinstance(mul, self.__class__):
            raise TypeError('Cannot multiply %s by %s' % (self, mul))
        if not isinstance(mul, numbers.Real):
            raise TypeError('Cannot multiply %s by %r' % (self, mul))
        return self.__class__(super(Duration, self).__mul__(mul))
    __rmul__ = __mul__

    def __truediv__(self, div):
        if isinstance(div, self.__class__):
            return super(Duration, self).__truediv__(div)
        return self * (1/div)

    def __floordiv__(self, div):
        res = float.__floordiv__(self, div)
        if not isinstance(div, self.__class__):
            # if div is not a duration, return a duration
            return self.__class__(res)
        # otherwise return a float
        return res

    def __mod__(self, div):
        return self.__class__(float.__mod__(self, div))

    def __rmod__(self, number):
        if number != 0:
            raise ArithmeticError("Cannot perform modulu of %s by %s" % (number, self))
        return 0

    def __neg__(self):
        return -1*self

    def rounddown(self, unit):
        return (self // unit) * unit

    def roundup(self, unit):
        if 0 == self % unit:
            return self
        else:
            return (self // unit + 1) * unit

    def round(self, unit):
        if self % unit > unit / 2:
            add = unit
        else:
            add = 0
        return self.rounddown(unit) + add

    def ceildiv(self, other):
        return self.roundup(other) / other

    def randrange(self, divisor, start=None):
        'returns a random size below self that divides by divisor'
        if start is None:
            start = self.__class__(0)
        start = start.ceildiv(divisor)
        stop = self.ceildiv(divisor)
        return random.randrange(start=start, stop=stop) * divisor

    def __add__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(float.__add__(self, other))
        elif isinstance(other, str):
            return self + self.__class__(other)
        else:
            return self.__class__(super(Duration, self).__add__(other))
    __radd__ = __add__

    def __format__(self, spec):
        width, precision, name = re.match(r"(?:(\d+))?(?:\.(\d*))?(\w+)?", spec).groups()
        if name == "f":
            return super(Duration, self).__format__(spec)
        dir = "<"
        if name == "text":
            from easypy.humanize import time_duration
            ret = time_duration(self)
        elif name == "ago":
            from easypy.humanize import time_duration
            ret = time_duration(self, ago=True)
        else:
            ret = self.render(unit=name, precision=precision)
            dir = ">"
        return "{:{dir}{width}}".format(ret, width=width or "", dir=dir)

    # don't override, so it can be converted to json properly
    # def __str__(self):

    def render(self, unit=None, precision=None):
        if self == NEVER:
            return "Eternity"

        if not unit:
            for unit_size in self.SORTED_UNITS:
                if unit_size <= abs(self):
                    break
            unit = self.UNIT_NAMES[unit_size]
        assert unit in self.NAMED_UNITS
        dd = hh = mm = 0
        ss = float(self)
        if unit not in ("ms", "s"):
            mm, ss = divmod(ss, 60)
        if unit != "m":
            hh, mm = divmod(mm, 60)
        if unit != "h":
            dd, hh = divmod(hh, 24)

        if dd or unit == "d":
            return "%.0fd, %02.0f:%02.0fh" % (dd, hh, mm)
        elif hh or unit == "h":
            return "%02.0f:%02.0fh" % (hh, mm)
        elif mm or unit == "m":
            return "%02.0f:%02.0fm" % (mm, ss)
        elif unit in (None, "", "s"):
            return "{:.{precision}f}s".format(ss, precision=precision if precision is not None else 0 if ss.is_integer() else 1)
        else:
            ms = ss * 1000
            return "{:.{precision}f}ms".format(ms, precision=(precision or 1) if ms < 1 else (precision or 0))
        assert False, "This Should not happen"

    __repr__ = render

    def _repr_pretty_(self, p, cycle):
        # used by IPython
        from easypy.colors import MAGENTA
        if cycle:
            p.text('Duration(...)')
            return
        p.text(MAGENTA(self))


byte, KiB, MiB, GiB, TiB, PiB, XiB = repeat(0, 7)  # fixing ugly import error messages and allowing autoimport
KB, MB, GB, TB, PB, XB = repeat(0, 6)

globals().update({name: DataSize(name) for name in DataSize.NAMED_UNITS.keys()})

NANOSECOND = Duration(1/1000000000)
MICROSECOND = Duration(1/1000000)
MILLISECOND = Duration(1/1000)
SECOND = Duration(1)
MINUTE = Duration(60 * SECOND)
HOUR = Duration(60 * MINUTE)
DAY = Duration(24 * HOUR)
WEEK = Duration(7 * DAY)
MONTH = Duration(31 * DAY)
YEAR = Duration(365 * DAY)
ETERNITY = NEVER = Duration("inf")

# ------


def range_compare(value, value_range):
    if isinstance(value_range, tuple):
        lower, upper = value_range
    else:
        upper = lower = value_range

    if upper is not None:
        if value > upper:
            return value-upper
    if lower is not None:
        if value < lower:
            return value-lower
    return 0


def to_data_size(size):
    from math import ceil
    sizes_table = dict(k=KiB, m=MiB, g=GiB, t=TiB, p=PiB)
    if isinstance(size, (int, float)):
        return size
    try:
        if size[-1].isalpha():
            return ceil(float(size[:-1])) * sizes_table[size[-1].lower()]
        else:
            return ceil(float(size))
    except (KeyError, ValueError):
            raise UnknownDataSizeError(size)



from copy import deepcopy

from .exceptions import TException
from .tokens import AUTO, MANDATORY
from .collections import ListCollection, iterable
from .bunch import Bunch, bunchify
from collections import OrderedDict


class InvalidFieldType(TException):
    template = 'Bad field type {field_type} - {reason}'


class NotFields(TException, AttributeError):
    template = '{field_names} are not fields of {typed_struct.__name__}'


class NotAField(TException, AttributeError):
    template = '{field_name} is not a field of {typed_struct.__name__}'


class NotAKey(TException, KeyError):
    template = '{field_name} is not a key of {typed_struct.__name__}'


class MissingField(TException, TypeError):
    template = '{typed_struct.__name__} is missing mandatory field {field.name}'


class FieldTypeMismatch(TException, TypeError):
    template = '{field.name} is {expected_type}, not {received_type}'


class FieldKeyTypeMismatch(TException, TypeError):
    template = 'keys of {field.name} are {expected_type}, not {received_type}'


class FieldCollectionTypeMismatch(TException, TypeError):
    template = '{field.name} is {field.collection_type.__name__} - not {received_type.__name__}'


class Field(object):
    """
    Define a single field in a TypedStruct

    :param type: The type of the field
        * Use [<type>] for a typechecked list
        * Use {<key_type>: <type>} for a typechecked dict
        * Use {str: <type>} for a typechecked Bunch
    :param default: Default value for the field
    :param preprocess: A function for preprocessing the values assigned to the field:
        * If the field can be converted from different types,
          should do the conversion and return the converted value.
        * If there are validation checks, should perform them and
          raise an exception if they fail.
        * Use the field.add_validation() and field.add_conversion()
          helpers to modify the preprocess function without
          manually setting this parameter.
        * Must return a value of the field's type.
    :param meta: Custom metadata. Can be queried by tools that use reflection
                 on the TypedStruct object.
    """

    def __init__(self, type, *, default=MANDATORY, repr=AUTO, hash=AUTO, preprocess=None, meta=Bunch()):
        if isinstance(type, list):
            if len(type) != 1:
                raise InvalidFieldType(field_type=type,
                                       reason='Collection-style fields must be defined with a single item collection')
            type, = type
            self.collection_type = TypedList
            self.key_type = None
        elif isinstance(type, dict):
            if len(type) != 1:
                raise InvalidFieldType(field_type=type,
                                       reason='Collection-style fields must be defined with a single item collection')
            (self.key_type, type), = type.items()
            if self.key_type is str:
                self.collection_type = TypedBunch
            else:
                self.collection_type = TypedDict
        else:
            self.collection_type = None
            self.key_type = None

        import builtins
        if not isinstance(type, builtins.type):
            raise InvalidFieldType(field_type=type,
                                   reason='Not a type')
        self.type = type

        self.default = default
        """
        A default value for the field

        If not set, the field will be ``MANDATORY``.

        >>> class Foo(TypedStruct):
        ...    a = int
        ...    a.default = 12
        >>> Foo()
        Foo(a=12)
        """

        self.repr = repr
        """
        How to represent the field when converting the typed struct to a string.

        >>> from easypy.typed_struct import TypedStruct
        >>> class Foo(TypedStruct):
        >>>     a = int
        >>>     a.repr = lambda n: '0x%X' % n  # show a as hex
        >>>     b = int
        >>>     b.repr = False  # don't show b in the representation
        >>> Foo(a=11, b=12)
        Foo(a=0xB)
        """

        self.hash = hash
        """
        How to hash the field in when creating an hash for the typed struct.

        >>> from easypy.typed_struct import TypedStruct
        >>> class Foo(TypedStruct):
        >>>     a = int
        >>>     a.hash = lambda a: a % 2  # a's hash will only be it's parity
        >>>     b = int
        >>>     b.hash = False  # b will not be considered in the hash
        """

        # NOTE: _validate_type() will be also be called in _process_new_value()
        # in case people implement their own preprocess function, but we still
        # want to call it before the validatiors added with add_validation() so
        # that they won't throw a generic TypeError that will prevent the more
        # specific FieldTypeMismatch.
        self.preprocess = preprocess or self._validate_type
        self.meta = Bunch(meta)
        self.name = None

        if issubclass(self.type, TypedStruct):
            if self.default is MANDATORY:
                try:
                    self.default = self.type()
                except MissingField:
                    pass  # keep it mandatory

            orig_preprocess = self.preprocess

            def preprocess(obj):
                if isinstance(obj, dict) and not isinstance(obj, TypedStruct):
                    obj = self.type.from_dict(obj)
                return orig_preprocess(obj)
            self.preprocess = preprocess

    def add_validation(self, predicate, ex_type, *ex_args, **ex_kwargs):
        """
        Add a validation on values assigned to the field

        :param predicate: The validation function. Should return True if the value is OK.
        :param ex_type: The exception to raise if the predicate returns False on a value.
        :param ex_args,ex_kwargs: Arguments for the exception.

        Calling this method will modify the preprocess function.

        >>> class Foo(TypedStruct):
        ...     a = int
        ...     a.add_validation(lambda value: value < 10, ValueError, 'value for `a` is too big')
        >>> Foo(a=5)
        Foo(a=5)
        >>> Foo(a=15)
        Traceback (most recent call last):
        ValueError: value for `a` is too big
        """
        orig_preprocess = self.preprocess

        def new_preprocess(obj):
            obj = orig_preprocess(obj)
            if not predicate(obj):
                raise ex_type(*ex_args, **ex_kwargs)
            return obj
        self.preprocess = new_preprocess

    def add_conversion(self, predicate, conversion):
        """
        Add a conversion of values assigned to the field

        :param predicate: Only perform the conversion if this returns True. If
                          the predicate is a type, the conversion will be
                          performed if the assigned value is of that type.
        :param conversion: A conversion function. The returned value must be of
                           the type of the field.

        Calling this method will modify the preprocess function.

        >>> class Foo(TypedStruct):
        ...     a = int
        ...     a.add_conversion(str, int)
        ...     a.add_conversion(list, len)  # passing a list will set `a` to the number of items
        >>> Foo(a=1)
        Foo(a=1)
        >>> Foo(a='2')
        Foo(a=2)
        >>> Foo(a=[10, 20, 30])  # the list has 3 items
        Foo(a=3)
        """
        if isinstance(predicate, type):
            typ = predicate

            def predicate(obj):
                return isinstance(obj, typ)

        orig_preprocess = self.preprocess

        def new_preprocess(obj):
            if predicate(obj):
                obj = conversion(obj)
                assert isinstance(obj, self.type), 'Conversion %s converted to %s - expected %s' % (
                    conversion, type(obj), self.type)
            return orig_preprocess(obj)
        self.preprocess = new_preprocess

    def convertible_from(self, *predicates, conversion=AUTO):
        """
        Add similar conversions from multiple types

        :param predicates: Only perform the conversion if any of them return true.
                           If a predicate is a type, the conversion will be performed
                           if the assigned value is of that type.
        :param conversion: A conversion function. The returned value must be of
                           the type of the field. Defaults to the field's type.

        Calling this method will modify the preprocess function.

        >>> class Foo(TypedStruct):
        ...     a = int
        ...     a.convertible_from(str, float)
        >>> Foo(a=1)
        Foo(a=1)
        >>> Foo(a='2')
        Foo(a=2)
        >>> Foo(a=3.0)
        Foo(a=3)
        """
        if conversion is AUTO:
            conversion = self.type

        for predicate in predicates:
            self.add_conversion(predicate, conversion)

    def _validate_type(self, value):
        if not isinstance(value, self.type):
            raise FieldTypeMismatch(field=self,
                                    expected_type=self.type,
                                    received_type=type(value)) from None
        return value

    def _named(self, name):
        assert self.name is None, '%s is already named' % self
        self.name = name
        return self

    def __get__(self, obj, _=None):
        if obj is None:
            return self
        return obj[self.name]

    def __set__(self, obj, value):
        if self.collection_type is not None:
            try:
                existing = obj[self.name]
            except KeyError:  # First time - called from TypedStruct c'tor
                assert isinstance(value, self.collection_type), '%s is not a %s' % (value, self.collection_type)
            else:  # Collection already exists - let it handle the assignment
                if existing is not value:
                    existing._set(value)
                return

        else:
            value = self._process_new_value(value)

        return super(TypedStruct, obj).__setitem__(self.name, value)

    def _process_new_value(self, value):
        value = self.preprocess(value)
        self._validate_type(value)
        return value

    def to_parameter(self):
        from inspect import Parameter

        return Parameter(
            name=self.name,
            kind=Parameter.KEYWORD_ONLY,
            default=Parameter.empty if self.default is MANDATORY else self.default,
            annotation=self.type)

    def __repr__(self):
        return 'Field<%s>' % self.to_parameter()

    def _clone(self):
        return type(self)(
            type=self.type,
            default=self.default,
            preprocess=self.preprocess,
            meta=deepcopy(self.meta))


class TypedCollection(object):
    def __init__(self, owner, field):
        self._owner = owner
        self._field = field
        super().__init__()


class TypedList(TypedCollection, ListCollection):
    def _set(self, values):
        if not iterable(values):
            raise FieldCollectionTypeMismatch(field=self._field, received_type=type(values))
        # Must eagerly verify them all before clearing the list
        values = [self._field._process_new_value(value) for value in values]
        self.clear()
        # Use super to avoid calling _process_new_value twice
        super().extend(values)

    def __setitem__(self, index, value):
        value = self._field._process_new_value(value)
        return super().__setitem__(index, value)

    def append(self, value):
        value = self._field._process_new_value(value)
        return super().append(value)

    def insert(self, index, value):
        value = self._field._process_new_value(value)
        return super().insert(index, value)

    def __iadd__(self, other):
        other = map(self._field._process_new_value, other)
        return super().__iadd__(other)

    def extend(self, iterable):
        iterable = map(self._field._process_new_value, iterable)
        return super().extend(iterable)

    # TODO: We may decide to override these methods in the future. They don't
    # need to validate members, but do change the collection - so if we want to
    # add events we need to override them:
    #       __delitem__(self, key)
    #       __imul__(self, other)
    #       clear(self)
    #       remove(self, value)
    #       reverse(self)
    #       sort(self, key=None, reverse=False)
    #       pop(self, index=None)


class TypedDict(TypedCollection, dict):
    def _set(self, values):
        if not isinstance(values, dict):
            raise FieldCollectionTypeMismatch(field=self._field, received_type=type(values))
        # Must eagerly verify them all before clearing the list
        values = {self._process_new_key(k): self._field._process_new_value(v) for k, v in values.items()}
        self.clear()
        # # Use super to avoid calling _process_new_value twice
        super().update(values)

    def _process_new_key(self, key):
        if not isinstance(key, self._field.key_type):
            raise FieldKeyTypeMismatch(field=self._field,
                                       expected_type=self._field.key_type,
                                       received_type=type(key))
        return key

    def __setitem__(self, key, value):
        key = self._process_new_key(key)
        value = self._field._process_new_value(value)
        return super().__setitem__(key, value)

    def setdefault(self, key, default=None):
        key = self._process_new_key(key)
        default = self._field._process_new_value(default)
        return super().setdefault(key, default)

    def update(self, dct={}, **kwargs):
        try:
            items = dct.items
        except AttributeError:
            items = dct
        else:
            items = items()

        for k, v in items:
            self[k] = v

        for k, v in kwargs.items():
            self[k] = v

    # TODO: We may decide to override these methods in the future. They don't
    # need to validate members, but do change the collection - so if we want to
    # add events we need to override them:
    #       __delitem__(self, key)
    #       clear(self)
    #       pop(k[,d])
    #       popitem(self)


class TypedBunch(TypedDict, Bunch):
    def __init__(self, owner, field):
        # Hack around Bunch's __getattr__()
        self.__dict__.update(_field=field, _owner=owner)


TypedBunch.__name__ = 'Bunch'  # make it look like a Bunch when formatted


class TypedStructMeta(type):
    @classmethod
    def __prepare__(metacls, name, bases, **kwds):
        return _TypedStructDslDict(base for base in bases if isinstance(base, metacls))

    def __new__(mcs, name, bases, dct):
        fields = []
        for k, v in dct.items():
            if isinstance(v, Field):
                fields.append(v)
        dct['_fields'] = fields
        if '__init__' not in dct:
            from inspect import Signature

            def __init__(self, **kwargs):
                super(cls, self).__init__(**kwargs)
            __init__.__signature__ = Signature(parameters=[field.to_parameter() for field in fields])
            dct['__init__'] = __init__

        try:
            if TypedStruct in bases and '__doc__' not in dct:
                # If we don't do this, IPython will display the TypedStruct docstring
                dct['__doc__'] = ""
        except NameError:  # TypedStruct itself is defined here
            pass

        cls = super().__new__(mcs, name, bases, dct)
        return cls


class _TypedStructDslDict(OrderedDict):
    def __init__(self, bases):
        super().__init__()
        for base in bases:
            for field in base._fields:
                super().__setitem__(field.name, field._clone()._named(field.name))

    def __setitem__(self, name, value):
        if isinstance(value, Field):
            value = value._named(name)
        else:
            try:
                value = Field(value)._named(name)
            except InvalidFieldType:
                pass
        return super().__setitem__(name, value)


class TypedStruct(dict, metaclass=TypedStructMeta):
    """
    Define typechecked-at-runtime classes.

    * Define fields with types, defaults values, validators etc.
    * Nesting of TypedStructs.
    * Typechecked collections.

    Example::

        import easypy.typed_struct as ts

        class Foo(ts.TypedStruct):
            a = ts.Field(int, default=14)
            b = ts.Field([str], default=['1', '2', '3'])  # list of strings

            # Shorter style
            c = bool
            d = [float]

    After assignment, fields are automatically converted to ``Field`` - so you
    can use the field definition helpers on them::

        from easypy.typed_struct import TypedStruct

        class Foo(TypedStruct):
            a = int
            a.default = 20
            a.convertible_from(str, float)
    """

    @classmethod
    def _get_field(cls, name, error_variant):
        field = getattr(cls, name, None)
        if not isinstance(field, Field):
            raise error_variant(typed_struct=cls,
                                field_name=name)
        return field

    @classmethod
    def from_dict(cls, dct):
        return cls(**dct)

    def items(self):
        for field in self._fields:
            yield field.name, field.__get__(self)

    def to_dict(self):
        return {name: value.to_dict() if isinstance(value, TypedStruct) else value
                for name, value in self.items()}

    def to_bunch(self):
        return bunchify(self)

    def __init__(self, **kwargs):
        for field in self._fields:
            if field.collection_type is not None:
                value = field.collection_type(self, field)
                try:
                    raw_value = kwargs.pop(field.name)
                except KeyError:
                    if field.default is not MANDATORY:
                        value._set(deepcopy(field.default))
                else:
                    value._set(raw_value)
            else:
                try:
                    value = kwargs.pop(field.name)
                except KeyError:
                    if field.default is MANDATORY:
                        raise MissingField(typed_struct=type(self),
                                           field=field) from None
                    value = deepcopy(field.default)

            field.__set__(self, value)
        if kwargs:
            raise NotFields(typed_struct=type(self), field_names=', '.join(kwargs.keys()))

    def __delitem__(self, key):
        assert False, 'deletion is not allowed'

    def __setitem__(self, key, value):
        self._get_field(key, NotAKey).__set__(self, value)

    def __setattr__(self, name, value):
        self._get_field(name, NotAField).__set__(self, value)

    def __gen_for_repr(self):
        from .humanize import _ReprAsString
        for field in self._fields:
            if field.repr is False:
                continue  # skip this field
            elif field.repr is AUTO:
                yield field.name, field.__get__(self)
            else:
                assert callable(field.repr), '%s.repr must be False, AUTO or a callable' % (field,)
                yield field.name, _ReprAsString(field.repr(field.__get__(self)))

    def __repr__(self):
        # return '%s(%s)' % (type(self).__name__, ', '.join(gen_parts()))
        return '%s(%s)' % (type(self).__name__, ', '.join('%s=%r' % pair for pair in self.__gen_for_repr()))

    def _repr_pretty_(self, p, cycle):
        from easypy.humanize import ipython_fields_repr
        return ipython_fields_repr(type(self).__name__, self.__gen_for_repr(), p, cycle)

    def __hash__(self):
        def gen_parts():
            yield type(self)
            for field in self._fields:
                if field.hash is False:
                    continue  # skip this field
                elif field.hash is AUTO:
                    yield field.__get__(self)
                else:
                    assert callable(field.hash), '%s.hash must be False, AUTO or a callable' % (field,)
                    yield field.hash(field.__get__(self))
        return hash(tuple(gen_parts()))

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return super().__eq__(other)

    def __ne__(self, other):
        return not self == other



from easypy.meta import EasyMeta, GetAllSubclasses


def test_easy_meta_before_cls_init():

    class FooMaker(metaclass=EasyMeta):
        @EasyMeta.Hook
        def before_subclass_init(name, bases, dct):
            dct[name] = "foo"

    class BarMaker(metaclass=EasyMeta):
        @EasyMeta.Hook
        def before_subclass_init(name, bases, dct):
            dct[name] = "bar"

    class Foo(FooMaker):
        ...
    assert Foo.Foo == "foo"

    class Bar(BarMaker):
        ...
    assert Bar.Bar == "bar"

    class Baz1(FooMaker, BarMaker):
        ...
    assert Baz1.Baz1 == "bar"

    class Baz2(BarMaker, FooMaker):
        ...
    assert Baz2.Baz2 == "foo"


def test_easy_meta_after_cls_init():
    class Foo(metaclass=EasyMeta):
        @EasyMeta.Hook
        def after_subclass_init(cls):
            cls.foo_init = cls.__name__

    class Bar(metaclass=EasyMeta):
        @EasyMeta.Hook
        def after_subclass_init(cls):
            cls.bar_init = cls.__name__

    class Baz(Foo, Bar):
        @EasyMeta.Hook
        def after_subclass_init(cls):
            cls.baz_init = cls.__name__

    assert not hasattr(Foo, 'foo_init'), 'after_subclass_init declared in Foo invoked on Foo'
    assert not hasattr(Bar, 'bar_init'), 'after_subclass_init declared in Bar invoked on Bar'

    assert Baz.foo_init == 'Baz'
    assert Baz.bar_init == 'Baz'
    assert not hasattr(Baz, 'baz_init'), 'after_subclass_init declared in Baz invoked on Baz'


def test_easy_meta_get_all_subclasses():
    class Foo(GetAllSubclasses):
        pass

    class Bar(Foo):
        pass

    class Baz(Foo):
        pass

    class Qux(Bar):
        pass

    assert set(Foo.get_all_subclasses()) == {Bar, Baz, Qux}
    assert set(Bar.get_all_subclasses()) == {Qux}
    assert set(Baz.get_all_subclasses()) == set()
    assert set(Qux.get_all_subclasses()) == set()


def test_easy_meta_multi_inheritance():

    class A(metaclass=EasyMeta):
        @EasyMeta.Hook
        def before_subclass_init(name, bases, dct):
            dct.setdefault('name', []).append("A")

    class B(metaclass=EasyMeta):
        @EasyMeta.Hook
        def before_subclass_init(name, bases, dct):
            dct.setdefault('name', []).append("B")

    class AA(A):
        @EasyMeta.Hook
        def before_subclass_init(name, bases, dct):
            dct.setdefault('name', []).append("AA")

    class BB(B):
        @EasyMeta.Hook
        def before_subclass_init(name, bases, dct):
            dct.setdefault('name', []).append("BB")

    class A_B(A, B): ...
    class AA_BB(AA, BB): ...
    class A_BB(A, BB): ...
    class B_AA(B, AA): ...

    # reminder - hooks aren't invoked on the classes on which they're defined, only subclasses
    assert not hasattr(A, "name")
    assert not hasattr(B, "name")
    assert AA.name == ["A"]
    assert BB.name == ["B"]
    assert A_B.name == ["A", "B"]
    assert AA_BB.name == ["A", "AA", "B", "BB"]
    assert A_BB.name == ["A", "B", "BB"]



import pytest
from easypy.deprecation import deprecated_arguments


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_deprecated_arguments():
    @deprecated_arguments(foo='bar')
    def func(bar):
        return 'bar is %s' % (bar,)

    assert func(1) == func(foo=1) == func(bar=1) == 'bar is 1'

    with pytest.raises(TypeError):
        func(foo=1, bar=2)

    with pytest.raises(TypeError):
        func(1, foo=2)



#!/usr/bin/env python

"""
This module allows rendering colored text in the console using a custom markup language::

    >> print(colorize("This will render as BLUE<<blue>>, and this will be YELLOW(BLUE)@{yellow-on-blue}@"))

"""

import re
import functools
from collections import namedtuple

ColorSpec = namedtuple("ColorSpec", "color light".split())


color_names = [
    'dark_gray',  # 0
    'red',        # 1
    'green',      # 2
    'yellow',     # 3
    'blue',       # 4
    'magenta',    # 5
    'cyan',       # 6
    'white',      # 7
]

COLORS = dict(
    (name, ColorSpec(color=color_num, light=True))
    for color_num, name in enumerate(color_names)
)

for orig in sorted(COLORS):
    COLORS['dark_' + orig] = ColorSpec(COLORS[orig].color, light=False)

for orig, dark in [('dark_gray', 'black'),
                   ('white', 'gray')]:
    COLORS[dark] = ColorSpec(COLORS[orig].color, light=False)


ANSI_BEGIN = "\x1b"
ANSI_END = ANSI_BEGIN + "[0m"


# this regex is used to split text into ANSI-colored/uncolored text blocks
RE_FIND_ANSI_COLORING = re.compile(r"({0}.*?{0}\[0m)".format(ANSI_BEGIN))

# this regex is used to tokenize an ANSI-colored string
RE_PARSE_ANSI_COLORING = re.compile(
    r"(?P<underline>{0}\[4m)"
    r"?{0}\[(?P<light>[01]\d*)"
    r"(?P<color>(?:;[1-9]\d*)*)m"
    r"(?P<text>.*?)"
    r"{0}\[0m".format(ANSI_BEGIN))


# this regex is used to split text into markup/non-markup text blocks
RE_FIND_COLOR_MARKUP = re.compile(
    r"(?ms)("
    r"[A-Z_]+(?:\([^\)]+\))?"
    r"(?:"
    r"(?:\<\<).*?(?:\>\>)|"
    r"(?:\@\{).*?(?:\}\@)|"
    r"(?:\@\[).*?(?:\]\@)"
    "))")

# this regex is used to parse the color markup into a foreground color, optional background, and the text itself.
# the text can be enclosed either by '<<..>>' or '@[...]@'
RE_PARSE_COLOR_MARKUP = re.compile(
    r"(?ms)"
    r"([A-Z_]+(?:\([^\)]+\))?)"  # group 0: the coloring
    r"(?:"
    r"\<\<(.*?)\>\>|"            # group 1: first trap for text <<...>>
    r"\@\{(.*?)\}\@|"            # group 2: second trap for text @{...}@
    r"\@\[(.*?)\]\@"             # group 3: second trap for text @[...]@
    ")")


class Colorized(str):
    """
    A string with coloring mark-up that retains string operation without interfering with the markup.
    For example::

        >> len(Colorized("RED<<red>>"))
        3

        >> Colorized("RED<<Red>>").lower()
        "RED<<red>>"
    """

    class Token(str):

        def raw(self):
            return self

        def copy(self, text):
            return self.__class__(text)

        def __getslice__(self, start, stop):
            return self[start:stop:]

        def __getitem__(self, *args):
            return self.copy(str.__getitem__(self, *args))

        def __iter__(self):
            for c in str.__str__(self):
                yield self.copy(c)

    class ColoredToken(Token):

        def __new__(cls, text, colorizer_name):
            self = str.__new__(cls, text)
            self.__name = colorizer_name
            self.__colored = Colorizer.from_markup(self.__name)(text)
            return self

        def __str__(self):
            return self.__colored

        def copy(self, text):
            return self.__class__(text, self.__name)

        def raw(self):
            return "%s<<%s>>" % (self.__name, str.__str__(self))

        def __repr__(self):
            return repr(self.raw())

    def __new__(cls, text):
        text = uncolored(text, markup=False)  # remove exiting ANSI colors
        self = str.__new__(cls, text)
        self.tokens = []
        for part in RE_FIND_COLOR_MARKUP.split(text):
            if not part:
                continue
            match = RE_PARSE_COLOR_MARKUP.match(part)
            if match:
                stl, *parts = match.groups()
                stl = stl.strip("_")
                part = next(filter(None, parts), "")  # default to "", in case there's no text in the token
                if part:
                    for l in part.splitlines():
                        self.tokens.append(self.ColoredToken(l, stl))
                        self.tokens.append(self.Token("\n"))
                    if not part.endswith("\n"):
                        del self.tokens[-1]
            else:
                self.tokens.append(self.Token(part))
        self.uncolored = "".join(str.__str__(token) for token in self.tokens)
        self.colored = "".join(str(token) for token in self.tokens)
        return self

    def raw(self):
        return str.__str__(self)

    def len_delta(self):
        return len(self.colored) - len(self.uncolored)

    def __str__(self):
        return self.colored

    def withuncolored(func):
        def inner(self, *args):
            return func(self.uncolored, *args)
        return inner

    __len__ = withuncolored(len)

    count = withuncolored(str.count)
    endswith = withuncolored(str.endswith)
    find = withuncolored(str.find)
    index = withuncolored(str.index)
    isalnum = withuncolored(str.isalnum)
    isalpha = withuncolored(str.isalpha)
    isdigit = withuncolored(str.isdigit)
    islower = withuncolored(str.islower)
    isspace = withuncolored(str.isspace)
    istitle = withuncolored(str.istitle)
    isupper = withuncolored(str.isupper)
    rfind = withuncolored(str.rfind)
    rindex = withuncolored(str.rindex)

    def withcolored(func):
        def inner(self, *args):
            return self.__class__("".join(t.copy(func(t, *args)).raw() for t in self.tokens if t))
        return inner

    capitalize = withcolored(str.capitalize)
    expandtabs = withcolored(str.expandtabs)
    lower = withcolored(str.lower)
    replace = withcolored(str.replace)

    # decode = withcolored(str.decode)
    # encode = withcolored(str.encode)
    swapcase = withcolored(str.swapcase)
    title = withcolored(str.title)
    upper = withcolored(str.upper)

    def __getitem__(self, idx):
        if isinstance(idx, slice) and idx.step is None:
            start = idx.start or 0
            stop = idx.stop or len(self)
            cursor = 0
            tokens = []
            for token in self.tokens:
                tokens.append(token[max(0, start - cursor):stop - cursor])
                cursor += len(token)
                if cursor > stop:
                    break
            return self.__class__("".join(t.raw() for t in tokens if t))

        tokens = [c for token in self.tokens for c in token].__getitem__(idx)
        return self.__class__("".join(t.raw() for t in tokens if t))

    def __add__(self, other):
        return self.__class__("".join(map(str.__str__, (self, other))))

    def __radd__(self, other):
        return self.__class__("".join(map(str.__str__, (other, self))))

    def __mod__(self, other):
        return self.__class__(self.raw() % other)

    def __rmod__(self, other):
        return self.__class__(other % self.raw())

    def format(self, *args, **kwargs):
        return self.__class__(self.raw().format(*args, **kwargs))

    def rjust(self, *args):
        padding = self.uncolored.rjust(*args)[:-len(self.uncolored)]
        return self.__class__(padding + self.raw())

    def ljust(self, *args):
        padding = self.uncolored.ljust(*args)[len(self.uncolored):]
        return self.__class__(self.raw() + padding)

    def center(self, *args):
        padded = self.uncolored.center(*args)
        return self.__class__(padded.replace(self.uncolored, self.raw()))

    def join(self, *args):
        return self.__class__(self.raw().join(*args))

    def _iter_parts(self, parts):
        last_cursor = 0
        for part in parts:
            pos = self.uncolored.find(part, last_cursor)
            yield self[pos:pos + len(part)]
            last_cursor = pos + len(part)

    def withiterparts(func):
        def inner(self, *args):
            return list(self._iter_parts(func(self.uncolored, *args)))
        return inner

    split = withiterparts(str.split)
    rsplit = withiterparts(str.rsplit)
    splitlines = withiterparts(str.splitlines)
    partition = withiterparts(str.partition)
    rpartition = withiterparts(str.rpartition)

    def withsingleiterparts(func):
        def inner(self, *args):
            return next(self._iter_parts([func(self.uncolored, *args)]))
        return inner

    strip = withsingleiterparts(str.strip)
    lstrip = withsingleiterparts(str.lstrip)
    rstrip = withsingleiterparts(str.rstrip)

    def zfill(self, *args):
        padding = self.uncolored.zfill(*args)[:-len(self.uncolored)]
        return self.__class__(padding + self.raw())

    @classmethod
    def from_ansi(cls, text):
        """
        Parse ANSI coloring and convert into Colorized text
        """
        parts = []
        for part in RE_FIND_ANSI_COLORING.split(text):
            if not part:
                continue
            match = RE_PARSE_ANSI_COLORING.match(part)
            if match:
                d = match.groupdict()
                fg = bg = None
                for color in d['color'].split(";"):
                    if not color:
                        continue
                    color = int(color)
                    if color >= 40:
                        bg = color_names[color - 40]
                    else:
                        fg = color_names[color - 30]

                color = fg.upper()
                if d['light'] in (None, "", "0"):
                    color = "DARK_{}".format(color)
                if bg:
                    color = "{}({})".format(color, bg.upper())
                part = d['text']
                if "<<" in part or ">>" in part:
                    sep = "@{", "}@"
                else:
                    sep = "<<", ">>"

                parts.append("{color}{sep[0]}{part}{sep[1]}".format(**locals()))
            else:
                parts.append(part)
        return cls("".join(parts))

    del withuncolored, withcolored


C = Colorized


def uncolored(text, ansi=True, markup=True):
    """
    Strip coloring markup and/or ANSI escape coloing from text
    """
    if ansi:
        text = re.sub(re.escape(ANSI_BEGIN) + '.+?m', "", text)
    if markup:
        text = RE_PARSE_COLOR_MARKUP.sub(lambda m: next(filter(None, m.groups()[1:])), text)
    return text


def colorize(text):
    """
    Colorize text according to markup
    """
    if not isinstance(text, str):
        text = str(text)

    def _subfunc(match_obj):
        colorizer = Colorizer.from_markup(match_obj.group(1))
        token = next(filter(None, match_obj.groups()[1:]), "")
        return colorizer(token) if token else ""

    return RE_PARSE_COLOR_MARKUP.sub(_subfunc, text)


class Colorizer():
    """
    A callable for styling text::

        >> warning = Colorizer('yellow', 'blue', underline=True, name='warning')
        >> warning("this is a warning")
        '\\x1b[4m\\x1b[1;44;33mthis is a warning\\x1b[0m'
    """

    COLORIZERS = {}

    def __new__(cls, color='white', background=None, underline=False, name=None):
        if name:
            name = name.lower()
            try:
                return cls.COLORIZERS[name]
            except KeyError:
                pass

        key = color, background, underline

        try:
            ret = cls.COLORIZERS[key]
        except KeyError:
            ret = cls.COLORIZERS[key] = super().__new__(cls)

        if name:
            cls.COLORIZERS[name] = ret

        return ret

    def __init__(self, color='white', background=None, underline=False, name=None):
        self.color = color
        self.background = background
        self.underline = underline
        self.name = name.lower() if name else None

        color_spec = COLORS[color]
        light = color_spec.light
        color = color_spec.color
        if not underline and not light:
            prefix_str = ANSI_BEGIN + "[0"
        elif not underline or not light:
            prefix_str = ANSI_BEGIN + "[%d" % (1 if light else 4)
        else:
            prefix_str = ANSI_BEGIN + "[4m" + ANSI_BEGIN + "[1"
        if background is not None:
            background = COLORS[background].color
            color_str = "%d;%d" % (background + 40, color + 30)
        else:
            color_str = "%d" % (color + 30,)

        self.fmt = "%s;%sm{}%s" % (prefix_str, color_str, ANSI_END)

    def __repr__(self):
        if self.name:
            return "{0.__class__.__name__}('{0.name}')".format(self)
        else:
            spec = self.color
            if self.background:
                spec = "{}/{}".format(spec, self.background)
            if self.underline:
                spec += ", underlined"
            return "{0.__class__.__name__}({1})".format(self, spec)

    def __call__(self, text):
        return self.fmt.format(text)

    # this regex is used to parse the color spec in a color markup into a foreground color and an optional background color
    RE_PARSE_COLOR_SPEC = re.compile(r"([\w]+)(?:\((.*)\))?")

    @classmethod
    def from_markup(cls, markup):
        try:
            return cls.COLORIZERS[markup.lower()]
        except KeyError:
            pass
        color, background = (c and c.lower().strip("_") for c in cls.RE_PARSE_COLOR_SPEC.match(markup).groups())
        if color not in COLORS:
            color = "white"
        if background not in COLORS:
            background = None
        return cls(color=color, background=background, name=markup)


def register_colorizers(**styles):
    """
    Install named colorizers::

        >> register_colorizers(
            warning="yellow",               # yellow foreground
            critical=("yellow", "red"),     # yellow on red
            url=("white", "blue", True),    # white on blue with underline
            )
    """

    ret = {}
    for name, style in styles.items():
        if isinstance(style, str):
            fg = style
            bg = None
            underline = False
        elif isinstance(style, tuple):
            fg, bg, *more = style
            underline = more[0] if more else False
        else:
            raise ValueError("Invalid style: %r (expecting an str or a fg/bg tuple)", (style,))
        ret[name] = Colorizer(name=name, color=fg, background=bg, underline=underline)
    return ret


globals().update((name.upper(), Colorizer(color=name, name=name)) for name in COLORS)


def main():
    """
    Colorize lines from stdin
    """
    import fileinput
    try:
        for line in fileinput.input(openhook=functools.partial(open, errors='replace')):
            print(colorize(line), end="", flush=True)
    except BrokenPipeError:
        pass


if __name__ == "__main__":
    main()



"""
Common initialization that needs to be called before any easypy module that deals with multithreading
"""

from uuid import uuid4, UUID
from weakref import WeakKeyDictionary


MAIN_UUID = UUID(int=0)
UUIDS_TREE = WeakKeyDictionary()
UUID_TO_IDENT = WeakKeyDictionary()
IDENT_TO_UUID = {}


def _set_thread_uuid(ident, parent_uuid=MAIN_UUID):
    uuid = uuid4()
    IDENT_TO_UUID[ident] = uuid
    UUIDS_TREE[uuid] = parent_uuid


def _set_main_uuid():
    import threading
    IDENT_TO_UUID[threading.main_thread().ident] = MAIN_UUID
    UUID_TO_IDENT[MAIN_UUID] = threading.main_thread().ident


_set_main_uuid()
